# Justpy Tutorial demo event_demo7 from docs/tutorial/handling_events.md
import justpy as jp

def button_click3(self, msg):
    self.num_clicked += 1
    # self.message.text = f'{self.text} clicked. Number of clicks: {self.num_clicked}'
    p = jp.P(text=f'{self.text} clicked. Number of clicks: {self.num_clicked}')
    self.message.add_component(p, 0)
    for button in msg.page.button_list:
        button.set_class('bg-blue-500')
        button.set_class('bg-blue-700', 'hover')
    self.set_class('bg-red-500')
    self.set_class('bg-red-700', 'hover')

def event_demo7():
    number_of_buttons = 25
    wp = jp.WebPage()
    button_div = jp.Div(classes='flex m-4 flex-wrap', a=wp)
    button_classes = 'w-32 mr-2 mb-2 bg-blue-500 hover:bg-blue-700 text-white font-bold py-2 px-4 rounded-full'
    message = jp.Div(classes='text-lg border m-2 p-2 overflow-auto h-64', a=wp)
    message.add(jp.P(text='No button clicked yet'))
    button_list = []
    for i in range(1, number_of_buttons + 1):
        b = jp.Button(text=f'Button {i}', a=button_div, classes=button_classes, click=button_click3)
        b.message = message
        b.num_clicked = 0
        button_list.append(b)
    wp.button_list = button_list   # The list will now be referenced by the WebPage instance attribute
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("event_demo7",event_demo7)



'''
Created on 2022-09-11

@author: wf
'''
from starlette.config import Config
import logging
import os
from ssl import PROTOCOL_SSLv23

config = Config("justpy.env")
DEBUG = config("DEBUG", cast=bool, default=True)
HOST = config("HOST", cast=str, default="127.0.0.1")
PORT = config("PORT", cast=int, default=8000)
CRASH = config("CRASH", cast=bool, default=False)
LATENCY = config("LATENCY", cast=int, default=0)
if LATENCY:
    print(f"Simulating latency of {LATENCY} ms")
HTML_404_PAGE = "justpy is sorry - that path doesn't exist"
MEMORY_DEBUG = config("MEMORY_DEBUG", cast=bool, default=False)
SESSIONS = config("SESSIONS", cast=bool, default=True)
SESSION_COOKIE_NAME = config("SESSION_COOKIE_NAME", cast=str, default="jp_token")
SECRET_KEY = config(
    "SECRET_KEY", default="$$$my_secret_string$$$"
)  # Make sure to change when deployed
LOGGING_LEVEL = config("LOGGING_LEVEL", default=logging.WARNING)
UVICORN_LOGGING_LEVEL = config("UVICORN_LOGGING_LEVEL", default="WARNING").lower()
COOKIE_MAX_AGE = config(
    "COOKIE_MAX_AGE", cast=int, default=60 * 60 * 24 * 7
)  # One week in seconds

SSL_VERSION = config("SSL_VERSION", default=PROTOCOL_SSLv23)
SSL_KEYFILE = config("SSL_KEYFILE", default="")
SSL_CERTFILE = config("SSL_CERTFILE", default="")


STATIC_DIRECTORY = config("STATIC_DIRECTORY", cast=str, default=os.getcwd())
STATIC_ROUTE = config("STATIC_MOUNT", cast=str, default="/static")
STATIC_NAME = config("STATIC_NAME", cast=str, default="static")
FAVICON = config(
    "FAVICON", cast=str, default=""
)  # If False gets value from https://elimintz.github.io/favicon.png
TAILWIND = config("TAILWIND", cast=bool, default=True)
QUASAR = config("QUASAR", cast=bool, default=False)
QUASAR_VERSION = config("QUASAR_VERSION", cast=str, default=None)
HIGHCHARTS = config("HIGHCHARTS", cast=bool, default=True)
KATEX = config("KATEX", cast=bool, default=False)
VEGA = config("VEGA", cast=bool, default=False)
BOKEH = config("BOKEH", cast=bool, default=False)
PLOTLY = config("PLOTLY", cast=bool, default=False)
DECKGL = config("DECKGL", cast=bool, default=False)
AGGRID = config("AGGRID", cast=bool, default=True)
AGGRID_ENTERPRISE = config("AGGRID_ENTERPRISE", cast=bool, default=False)
NO_INTERNET = config("NO_INTERNET", cast=bool, default=True)
FRONTEND_ENGINE_TYPE = config("FRONTEND_ENGINE_TYPE", cast=str, default="vue")



from starlette.applications import Starlette
from starlette.responses import HTMLResponse, JSONResponse, Response
from starlette.responses import PlainTextResponse
from starlette.endpoints import WebSocketEndpoint
from starlette.endpoints import HTTPEndpoint
from starlette.middleware import Middleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware
from starlette.staticfiles import StaticFiles
from justpy.htmlcomponents import *
from .chartcomponents import *
from .gridcomponents import *
from .quasarcomponents import *

from jpcore.justpy_app import cookie_signer, template_options, handle_event, JustpyApp,JustpyAjaxEndpoint
from jpcore.justpy_config import config, AGGRID, AGGRID_ENTERPRISE,BOKEH,COOKIE_MAX_AGE, CRASH
from jpcore.justpy_config import DEBUG,DECKGL, FAVICON, HIGHCHARTS,HOST,KATEX, LATENCY,LOGGING_LEVEL
from jpcore.justpy_config import MEMORY_DEBUG, NO_INTERNET, PLOTLY, PORT, SECRET_KEY, SESSION_COOKIE_NAME, SESSIONS
from jpcore.justpy_config import SSL_CERTFILE, SSL_KEYFILE, SSL_VERSION, STATIC_DIRECTORY,STATIC_NAME, STATIC_ROUTE
from jpcore.justpy_config import QUASAR, QUASAR_VERSION,TAILWIND, UVICORN_LOGGING_LEVEL, VEGA
JustPy.LOGGING_LEVEL = LOGGING_LEVEL
# from .misccomponents import *
from .pandas import *
from .routing import SetRoute
from jpcore.utilities import run_task, create_delayed_task
import uvicorn, logging, sys, os, traceback
import typing
#
# globals
#
# uvicorn Server
jp_server = None
current_module = sys.modules[__name__]
current_dir = os.path.dirname(current_module.__file__)
print(current_dir.replace("\\", "/"))
print(f"Module directory: {current_dir}, Application directory: {os.getcwd()}")

logging.basicConfig(level=LOGGING_LEVEL, format="%(levelname)s %(module)s: %(message)s")

# modify middleware handling according to deprecation
# https://github.com/encode/starlette/discussions/1762
middleware = [Middleware(GZipMiddleware)]
if SSL_KEYFILE and SSL_CERTFILE:
    middleware.append(Middleware(HTTPSRedirectMiddleware))
#@TODO     
# implement https://github.com/justpy-org/justpy/issues/535
#if SESSIONS:
#    middleware.append(Middleware(SessionMiddleware, secret_key=SECRET_KEY))
app = JustpyApp(middleware=middleware, debug=DEBUG)
app.mount(STATIC_ROUTE, StaticFiles(directory=STATIC_DIRECTORY), name=STATIC_NAME)
app.mount(
    "/templates", StaticFiles(directory=current_dir + "/templates"), name="templates"
)


def initial_func(_request):
    """
    Default func/endpoint to be called if none has been specified
    """
    wp = WebPage()
    Div(
        text="JustPy says: Page not found",
        classes="inline-block text-5xl m-3 p-3 text-white bg-blue-600",
        a=wp,
    )
    return wp


func_to_run = initial_func
startup_func = None


def server_error_func(request):
    wp = WebPage()
    Div(
        text="JustPy says: 500 - Server Error",
        classes="inline-block text-5xl m-3 p-3 text-white bg-red-600",
        a=wp,
    )
    return wp

@app.on_event("startup")
async def justpy_startup():
    WebPage.loop = asyncio.get_event_loop()
    JustPy.loop = WebPage.loop
    JustPy.STATIC_DIRECTORY = STATIC_DIRECTORY

    if startup_func:
        if inspect.iscoroutinefunction(startup_func):
            await startup_func()
        else:
            startup_func()
    protocol = "https" if SSL_KEYFILE else "http"
    print(f"JustPy ready to go on {protocol}://{HOST}:{PORT}")

    
@app.route("/zzz_justpy_ajax")
class AjaxEndpoint(JustpyAjaxEndpoint):
    """
    Justpy ajax handler
    """
    
@app.websocket_route("/")
class JustpyEvents(WebSocketEndpoint):

    socket_id = 0

    async def on_connect(self, websocket):
        await websocket.accept()
        websocket.id = JustpyEvents.socket_id
        websocket.open = True
        logging.debug(f"Websocket {JustpyEvents.socket_id} connected")
        JustpyEvents.socket_id += 1
        # Send back socket_id to page
        # await websocket.send_json({'type': 'websocket_update', 'data': websocket.id})
        WebPage.loop.create_task(
            websocket.send_json({"type": "websocket_update", "data": websocket.id})
        )

    async def on_receive(self, websocket, data):
        """
        Method to accept and act on data received from websocket
        """
        logging.debug("%s %s", f"Socket {websocket.id} data received:", data)
        data_dict = json.loads(data)
        msg_type = data_dict["type"]
        # data_dict['event_data']['type'] = msg_type
        if msg_type == "connect":
            # Initial message sent from browser after connection is established
            # WebPage.sockets is a dictionary of dictionaries
            # First dictionary key is page id
            # Second dictionary key is socket id
            page_key = data_dict["page_id"]
            websocket.page_id = page_key
            if page_key in WebPage.sockets:
                WebPage.sockets[page_key][websocket.id] = websocket
            else:
                WebPage.sockets[page_key] = {websocket.id: websocket}
            return
        if msg_type == "event" or msg_type == "page_event":
            # Message sent when an event occurs in the browser
            session_cookie = websocket.cookies.get(SESSION_COOKIE_NAME)
            if SESSIONS and session_cookie:
                session_id = cookie_signer.unsign(session_cookie).decode("utf-8")
                data_dict["event_data"]["session_id"] = session_id
            # await self._event(data_dict)
            data_dict["event_data"]["msg_type"] = msg_type
            page_event = True if msg_type == "page_event" else False
            WebPage.loop.create_task(
                handle_event(data_dict, com_type=0, page_event=page_event)
            )
            return
        if msg_type == "zzz_page_event":
            # Message sent when an event occurs in the browser
            session_cookie = websocket.cookies.get(SESSION_COOKIE_NAME)
            if SESSIONS and session_cookie:
                session_id = cookie_signer.unsign(session_cookie).decode("utf-8")
                data_dict["event_data"]["session_id"] = session_id
            data_dict["event_data"]["msg_type"] = msg_type
            WebPage.loop.create_task(
                handle_event(data_dict, com_type=0, page_event=True)
            )
            return

    async def on_disconnect(self, websocket, close_code):
        try:
            pid = websocket.page_id
        except:
            return
        websocket.open = False
        WebPage.sockets[pid].pop(websocket.id)
        if not WebPage.sockets[pid]:
            WebPage.sockets.pop(pid)
        await WebPage.instances[pid].on_disconnect(
            websocket
        )  # Run the specific page disconnect function
        if MEMORY_DEBUG:
            print("************************")
            print(
                "Elements: ",
                len(JustpyBaseComponent.instances),
                JustpyBaseComponent.instances,
            )
            print("WebPages: ", len(WebPage.instances), WebPage.instances)
            print("Sockets: ", len(WebPage.sockets), WebPage.sockets)
            import psutil
            process = psutil.Process(os.getpid())
            print(f"Memory used: {process.memory_info().rss:,}")
            print("************************")

def get_server():
    """
    Workaround for global variable jp_server not working as expected
    """
    return jp_server

def Route(path:str,wpfunc:typing.Callable):
    """
    legacy Route handling
    
    Args:
        path (str): the path of the route to add
        wpfunc(Callable): a WebPage returning function to be added
    """
    app.add_jproute(path,wpfunc)


def justpy(
    func=None,
    *,
    start_server: bool = True,
    websockets: bool = True,
    host: str = HOST,
    port: int = PORT,
    startup=None,
    init_server: bool = True,
    **kwargs,
):
    """

    The main justpy entry point

    Args:
        func: the callback to get the webpage
        start_server(bool): if True start the server
        websockets(bool): if True use websockets
        host(str): the host to start from e.g. localhost or 0.0.0.0 to listen on all interfaces
        port(int): the port to use for listening
        startup: a callback for the startup phase
        init_server(bool): if True construct the server
        kwargs: further keyword arguments

    """
    global jp_server, func_to_run, startup_func, HOST, PORT

    HOST = host
    PORT = port
    if func:
        func_to_run = func
    else:
        func_to_run = initial_func
    if startup:
        startup_func = startup
    if websockets:
        WebPage.use_websockets = True
    else:
        WebPage.use_websockets = False
    app.add_jproute("/", func_to_run)
    for k, v in kwargs.items():
        template_options[k.lower()] = v

    if init_server:
        if SSL_KEYFILE and SSL_CERTFILE:
            uvicorn_config = uvicorn.config.Config(
                app,
                host=host,
                port=port,
                log_level=UVICORN_LOGGING_LEVEL,
                proxy_headers=True,
                ssl_keyfile=SSL_KEYFILE,
                ssl_certfile=SSL_CERTFILE,
                ssl_version=SSL_VERSION,
            )
        else:
            uvicorn_config = uvicorn.config.Config(
                app, host=host, port=port, log_level=UVICORN_LOGGING_LEVEL
            )
        jp_server = uvicorn.Server(uvicorn_config)
        if start_server:
            jp_server.run()

    return func_to_run


def convert_dict_to_object(d):
    """
    convert the given dict to an object
    """
    obj = globals()[d["class_name"]]()
    for obj_prop in d["object_props"]:
        obj.add(convert_dict_to_object(obj_prop))
    # combine the dictionaries
    for k, v in {**d, **d["attrs"]}.items():
        if k != "id":
            obj.__dict__[k] = v
    return obj


def redirect(url:str)->WebPage:
    """
    redirect to the given url
    
    Args:
        url(str): the url to redirect to
        
    Returns:
        a WebPage with a single Div that hat the redirect
    """
    wp = WebPage()
    wp.add(Div())
    wp.redirect = url
    return wp



# Justpy Tutorial demo children_test from docs/reference/htmlcomponent.md
import justpy as jp

def children_test():
    wp = jp.WebPage()
    div_classes = 'm-2 p-2 bg-blue-500 text-white text-lg'
    span_classes = 'm-2 p-2 bg-blue-500 text-yellow-700 text-xl'
    jp.Div(children=[jp.Div(classes=div_classes, children=
                            [jp.Span(text='s1', classes=span_classes), jp.Span(text='s2', classes=span_classes)])
        ,jp.Div(text='d2', classes=div_classes), jp.Div(text='d3', classes=div_classes), jp.Div(text='d4', classes=div_classes)], a=wp)

    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("children_test",children_test)



"""
Created on 2022-09-07

@author: wf
"""
import argparse
import socket
import typing

class Demo(object):
    """
    Base class for justpy demos to allow selecting host and port from commandline
    """
    testmode = False

    def __init__(self, name: str, wp_func: typing.Callable, **kwargs):
        """
        Constructor

        Args:
            name(str): the name of the demo
            host(int): the port to run the demo at (default:8000)
            wp_func(callable): the webpage callback
            **kwargs: further keyword arguments
        """
        self.name=name
        self.wp_func=wp_func
        self.kwargs=kwargs
        # make sure Demo code may be tested without automatically starting
        if Demo.testmode:
            return
        parser = argparse.ArgumentParser(description=name)
        parser.add_argument("--host", default=socket.getfqdn())
        parser.add_argument("--port", type=int, default=8000)
        args = parser.parse_args()
        import justpy as jp

        jp.justpy(wp_func, host=args.host, port=args.port, **kwargs)



"""
Created on 2022-08-25

@author: wf
"""
import asyncio
from selenium.webdriver.common.by import By
import justpy as jp
from tests.base_selenium_test import BaseSeleniumTest

class TestClickDemoWithSelenium(BaseSeleniumTest):
    """
    testing actual browser behavior with selenium
    """

    async def setUp(self):
        await BaseSeleniumTest.setUp(self,port=8124)

    async def onDivClick(self, msg):
        """
        handle the click of the div
        """
        print(msg)
        self.clickCount += 1
        msg.target.text = f"I was clicked {self.clickCount} times"

    async def wp_to_test(self):
        """
        the example Webpage under test
        """
        wp = jp.WebPage(debug=True)
        self.clickCount = 0
        d = jp.Div(
            text="Not clicked yet",
            a=wp,
            classes="w-48 text-xl m-2 p-1 bg-blue-500 text-white",
        )
        d.on("click", self.onDivClick)
        d.additional_properties = [
            "screenX",
            "pageY",
            "altKey",
            "which",
            "movementX",
            "button",
            "buttons",
        ]
        return wp

    # @unittest.skipIf(BaseClienttest.inPublicCI(), "clickdemo switched off in PublicCI")
    async def testClickDemo(self):
        """
        this will actually start a firefox browser and the websocket reload dialog will appear
        """
        await self.server.start(self.wp_to_test)
        url = self.server.get_url("/")
        self.browser.get(url)
        await asyncio.sleep(self.server.sleep_time)
        divs = self.browser.find_elements(By.TAG_NAME, "div")
        # get the clickable div
        div = divs[1]
        self.assertEqual("Not clicked yet", div.text)
        for i in range(5):
            div.click()
            await asyncio.sleep(self.server.sleep_time)
            self.assertEqual(f"I was clicked {i+1} times", div.text)
        self.browser.close()
        await asyncio.sleep(self.server.sleep_time)
        await self.server.stop()


# Justpy Tutorial demo html_comps5 from docs/tutorial/html_components.md
import justpy as jp

def html_comps5():
    wp = jp.WebPage()
    for j in range(10):
        p = jp.P(text=f'אני אוהב לתכנת בפייתון', a=wp, contenteditable=True, classes='text-white bg-blue-500 hover:bg-blue-700 ml-1 p-1 w-1/2')
        p.dir = 'rtl'
        p.lang = 'he'
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("html_comps5",html_comps5)



# Justpy Tutorial demo input_test2 from docs/quasar_tutorial/quasar_components.md
import justpy as jp

def color_change(self, msg):
    self.div.style = f'color: {self.value}'

def input_test2():
    wp = jp.QuasarPage(data={'color': ''})
    in1 = jp.QInput(filled=True, style='width: 400px', a=wp, model=[wp, 'color'], classes="q-pa-md", input=color_change)
    j = jp.parse_html("""
        <q-icon name="colorize" class="cursor-pointer">
                <q-popup-proxy transition-show="scale" transition-hide="scale">
                  <q-color name="color_input"/>
                </q-popup-proxy>
              </q-icon>
        """)
    in1.add_scoped_slot('append', j)
    color_input = j.name_dict['color_input']
    color_input.model = [wp, 'color']
    color_input.on('change', color_change)
    in1.div = jp.Div(text='Change this text color using QInput above', classes="q-pa-md text-h4", a=wp)
    color_input.div = in1.div
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("input_test2",input_test2)



'''
Created on 2022-09-13

@author: wf
'''
from jpcore.tailwind import Tailwind

# @TODO refactor as per #528
class Component(Tailwind):
    """
    keep track of ids an instances
    """
    next_id = 1
    instances = {}    



# Justpy Tutorial demo plotly_test from docs/tutorial/plotly.md
import justpy as jp
import plotly.figure_factory as ff
import numpy as np


def create_chart(numpoints):
    x1 = np.random.randn(numpoints) - 2
    x2 = np.random.randn(numpoints)
    x3 = np.random.randn(numpoints) + 2
    # Group data together
    hist_data = [x1, x2, x3]
    group_labels = ['Group 1', 'Group 2', 'Group 3']
    # Create display with custom bin_size
    fig = ff.create_distplot(hist_data, group_labels, bin_size=[.1, .25, .5])
    fig.update_layout(width=500, height=600)
    return fig


def generate_data(self, msg):
    wp = msg.page
    wp.c.chart = create_chart(wp.numpoints)


def plotly_test(request):
    wp = jp.WebPage()
    wp.numpoints = 500
    jp.Button(text='Generate New Data', click=generate_data, a=wp, classes=jp.Styles.button_simple + ' m-2 p-2')
    wp.c = jp.PlotlyChart(chart=create_chart(wp.numpoints), a=wp, classes='border m-2 p-6', style='width: 600px')
    return wp


# initialize the demo
from  examples.basedemo import Demo
Demo ("plotly_test",plotly_test, PLOTLY=True)



# Justpy Tutorial demo hello_world2 from docs/tutorial/basic_concepts.md
import justpy as jp


def hello_world2():
    wp = jp.WebPage()
    for i in range(1, 11):
        jp.P(text=f"{i}) Hello World!", a=wp, style=f"font-size: {10*i}px")
    return wp


# initialize the demo
from examples.basedemo import Demo

Demo("hello_world2", hello_world2)



# https://tailwindcss.com


class Tailwind:
    """
    handle css classes to be attached to components
    """
    # TODO: Customization - https://tailwindcss.com/docs/configuration
    pseudo_classes = [
        "hover",
        "focus",
        "active",
        "group-hover",
        "focus-within",
        "first",
        "last",
        "odd",
        "even",
        "disabled",
        "visited",
        "sm",
        "md",
        "lg",
        "xl",
        "group-focus",
    ]

    tw_dict = {
        "container": ["container"],
        "screen_reader": ["sr-only", "not-sr-only"],
        "display": [
            "block",
            "inline-block",
            "inline",
            "flex",
            "inline-flex",
            "table",
            "table-row",
            "table-cell",
            "table-caption",
            "table-column",
            "table-column-group",
            "table-header-group",
            "table-row-group",
            "table-footer-group",
            "hidden",
            "grid",
            "inline-grid",
            "flow-root",
        ],
        "float": ["float-right", "float-left", "float-none", "clearfix"],
        "clear": ["clear-right", "clear-left", "clear-none", "clear-both"],
        "object_fit": [
            "object-contain",
            "object-cover",
            "object-fill",
            "object-none",
            "object-scale-down",
        ],
        "object_position": [
            "object-bottom",
            "object-center",
            "object-left",
            "object-left-bottom",
            "object-left-top",
            "object-right",
            "object-right-bottom",
            "object-right-top",
            "object-top",
        ],
        "overflow": [
            "overflow-auto",
            "overflow-hidden",
            "overflow-visible",
            "overflow-scroll",
            "overflow-x-auto",
            "overflow-y-auto",
            "overflow-x-hidden",
            "overflow-y-hidden",
            "overflow-x-visible",
            "overflow-y-visible",
            "overflow-x-scroll",
            "overflow-y-scroll",
            "scrolling-touch",
            "scrolling-auto",
        ],
        "position": ["static", "fixed", "absolute", "relative", "sticky"],
        "top_right_bottom_left": [
            "inset-0",
            "inset-y-0",
            "inset-x-0",
            "top-0",
            "right-0",
            "bottom-0",
            "left-0",
            "inset-auto",
            "inset-y-auto",
            "inset-x-auto",
            "top-auto",
            "bottom-auto",
            "left-auto",
            "right-auto",
        ],
        "visibility": ["visible", "invisible"],
        "z_index": ["z-0", "z-10", "z-20", "z-30", "z-40", "z-50", "z-auto"],
        "font_family": ["font-sans", "font-serif", "font-mono"],
        "font_size": [
            "text-xs",
            "text-sm",
            "text-base",
            "text-lg",
            "text-xl",
            "text-2xl",
            "text-3xl",
            "text-4xl",
            "text-5xl",
            "text-6xl",
        ],
        "font_smoothing": ["antialiased", "subpixel-antialiased"],
        "font_style": ["italic", "not-italic"],
        "font_weight": [
            "font-hairline",
            "font-thin",
            "font-light",
            "font-normal",
            "font-medium",
            "font-semibold",
            "font-bold",
            "font-extrabold",
            "font-black",
        ],
        "letter_spacing": [
            "tracking-tighter",
            "tracking-tight",
            "tracking-normal",
            "tracking-wide",
            "tracking-wider",
            "tracking-widest",
        ],
        "line_height": [
            "leading-none",
            "leading-tight",
            "leading-snug",
            "leading-normal",
            "leading-relaxed",
            "leading-loose",
            "leading-3",
            "leading-4",
            "leading-5",
            "leading-6",
            "leading-7",
            "leading-8",
            "leading-9",
            "leading-10",
        ],
        "list_style_type": ["list-none", "list-disc", "list-decimal"],
        "list_style_position": ["list-inside", "list-outside"],
        "text_align": ["text-left", "text-center", "text-right", "text-justify"],
        "text_color": [
            "text-transparent",
            "text-current",
            "text-black",
            "text-white",
            "text-gray-100",
            "text-gray-200",
            "text-gray-300",
            "text-gray-400",
            "text-gray-500",
            "text-gray-600",
            "text-gray-700",
            "text-gray-800",
            "text-gray-900",
            "text-red-100",
            "text-red-200",
            "text-red-300",
            "text-red-400",
            "text-red-500",
            "text-red-600",
            "text-red-700",
            "text-red-800",
            "text-red-900",
            "text-orange-100",
            "text-orange-200",
            "text-orange-300",
            "text-orange-400",
            "text-orange-500",
            "text-orange-600",
            "text-orange-700",
            "text-orange-800",
            "text-orange-900",
            "text-yellow-100",
            "text-yellow-200",
            "text-yellow-300",
            "text-yellow-400",
            "text-yellow-500",
            "text-yellow-600",
            "text-yellow-700",
            "text-yellow-800",
            "text-yellow-900",
            "text-green-100",
            "text-green-200",
            "text-green-300",
            "text-green-400",
            "text-green-500",
            "text-green-600",
            "text-green-700",
            "text-green-800",
            "text-green-900",
            "text-teal-100",
            "text-teal-200",
            "text-teal-300",
            "text-teal-400",
            "text-teal-500",
            "text-teal-600",
            "text-teal-700",
            "text-teal-800",
            "text-teal-900",
            "text-blue-100",
            "text-blue-200",
            "text-blue-300",
            "text-blue-400",
            "text-blue-500",
            "text-blue-600",
            "text-blue-700",
            "text-blue-800",
            "text-blue-900",
            "text-indigo-100",
            "text-indigo-200",
            "text-indigo-300",
            "text-indigo-400",
            "text-indigo-500",
            "text-indigo-600",
            "text-indigo-700",
            "text-indigo-800",
            "text-indigo-900",
            "text-purple-100",
            "text-purple-200",
            "text-purple-300",
            "text-purple-400",
            "text-purple-500",
            "text-purple-600",
            "text-purple-700",
            "text-purple-800",
            "text-purple-900",
            "text-pink-100",
            "text-pink-200",
            "text-pink-300",
            "text-pink-400",
            "text-pink-500",
            "text-pink-600",
            "text-pink-700",
            "text-pink-800",
            "text-pink-900",
        ],
        "text_decoration": ["underline", "line-through", "no-underline"],
        "text_transform": ["uppercase", "lowercase", "capitalize", "normal-case"],
        "vertical_align": [
            "align-baseline",
            "align-top",
            "align-middle",
            "align-bottom",
            "align-text-top",
            "align-text-bottom",
        ],
        "whitespace": [
            "whitespace-normal",
            "whitespace-no-wrap",
            "whitespace-pre",
            "whitespace-pre-line",
            "whitespace-pre-wrap",
        ],
        "word_break": ["break-normal", "break-words", "break-all", "truncate"],
        "background_attachment": ["bg-fixed", "bg-local", "bg-scroll"],
        "background_color": [
            "bg-transparent",
            "bg-current",
            "bg-black",
            "bg-white",
            "bg-gray-50",
            "bg-gray-100",
            "bg-gray-200",
            "bg-gray-300",
            "bg-gray-400",
            "bg-gray-500",
            "bg-gray-600",
            "bg-gray-700",
            "bg-gray-800",
            "bg-gray-900",
            "bg-red-100",
            "bg-red-200",
            "bg-red-300",
            "bg-red-400",
            "bg-red-500",
            "bg-red-600",
            "bg-red-700",
            "bg-red-800",
            "bg-red-900",
            "bg-orange-100",
            "bg-orange-200",
            "bg-orange-300",
            "bg-orange-400",
            "bg-orange-500",
            "bg-orange-600",
            "bg-orange-700",
            "bg-orange-800",
            "bg-orange-900",
            "bg-yellow-100",
            "bg-yellow-200",
            "bg-yellow-300",
            "bg-yellow-400",
            "bg-yellow-500",
            "bg-yellow-600",
            "bg-yellow-700",
            "bg-yellow-800",
            "bg-yellow-900",
            "bg-green-100",
            "bg-green-200",
            "bg-green-300",
            "bg-green-400",
            "bg-green-500",
            "bg-green-600",
            "bg-green-700",
            "bg-green-800",
            "bg-green-900",
            "bg-teal-100",
            "bg-teal-200",
            "bg-teal-300",
            "bg-teal-400",
            "bg-teal-500",
            "bg-teal-600",
            "bg-teal-700",
            "bg-teal-800",
            "bg-teal-900",
            "bg-blue-100",
            "bg-blue-200",
            "bg-blue-300",
            "bg-blue-400",
            "bg-blue-500",
            "bg-blue-600",
            "bg-blue-700",
            "bg-blue-800",
            "bg-blue-900",
            "bg-indigo-100",
            "bg-indigo-200",
            "bg-indigo-300",
            "bg-indigo-400",
            "bg-indigo-500",
            "bg-indigo-600",
            "bg-indigo-700",
            "bg-indigo-800",
            "bg-indigo-900",
            "bg-purple-100",
            "bg-purple-200",
            "bg-purple-300",
            "bg-purple-400",
            "bg-purple-500",
            "bg-purple-600",
            "bg-purple-700",
            "bg-purple-800",
            "bg-purple-900",
            "bg-pink-100",
            "bg-pink-200",
            "bg-pink-300",
            "bg-pink-400",
            "bg-pink-500",
            "bg-pink-600",
            "bg-pink-700",
            "bg-pink-800",
            "bg-pink-900",
        ],
        "background_position": [
            "bg-bottom",
            "bg-center",
            "bg-left",
            "bg-left-bottom",
            "bg-left-top",
            "bg-right",
            "bg-right-bottom",
            "bg-right-top",
            "bg-top",
        ],
        "background_repeat": [
            "bg-repeat",
            "bg-no-repeat",
            "bg-repeat-x",
            "bg-repeat-y",
            "bg-repeat-round",
            "bg-repeat-space",
        ],
        "background_size": ["bg-auto", "bg-cover", "bg-contain"],
        "background_opacity": [
            "bg-opacity-0",
            "bg-opacity-25",
            "bg-opacity-50",
            "bg-opacity-75",
            "bg-opacity-100",
        ],
        "text_opacity": [
            "text-opacity-0",
            "text-opacity-25",
            "text-opacity-50",
            "text-opacity-75",
            "text-opacity-100",
        ],
        "placeholder_opacity": [
            "placeholder-opacity-0",
            "placeholder-opacity-25",
            "placeholder-opacity-50",
            "placeholder-opacity-75",
            "placeholder-opacity-100",
        ],
        "border_opacity": [
            "border-opacity-0",
            "border-opacity-25",
            "border-opacity-50",
            "border-opacity-75",
            "border-opacity-100",
        ],
        "border_color": [
            "border-transparent",
            "border-current",
            "border-black",
            "border-white",
            "border-gray-100",
            "border-gray-200",
            "border-gray-300",
            "border-gray-400",
            "border-gray-500",
            "border-gray-600",
            "border-gray-700",
            "border-gray-800",
            "border-gray-900",
            "border-red-100",
            "border-red-200",
            "border-red-300",
            "border-red-400",
            "border-red-500",
            "border-red-600",
            "border-red-700",
            "border-red-800",
            "border-red-900",
            "border-orange-100",
            "border-orange-200",
            "border-orange-300",
            "border-orange-400",
            "border-orange-500",
            "border-orange-600",
            "border-orange-700",
            "border-orange-800",
            "border-orange-900",
            "border-yellow-100",
            "border-yellow-200",
            "border-yellow-300",
            "border-yellow-400",
            "border-yellow-500",
            "border-yellow-600",
            "border-yellow-700",
            "border-yellow-800",
            "border-yellow-900",
            "border-green-100",
            "border-green-200",
            "border-green-300",
            "border-green-400",
            "border-green-500",
            "border-green-600",
            "border-green-700",
            "border-green-800",
            "border-green-900",
            "border-teal-100",
            "border-teal-200",
            "border-teal-300",
            "border-teal-400",
            "border-teal-500",
            "border-teal-600",
            "border-teal-700",
            "border-teal-800",
            "border-teal-900",
            "border-blue-100",
            "border-blue-200",
            "border-blue-300",
            "border-blue-400",
            "border-blue-500",
            "border-blue-600",
            "border-blue-700",
            "border-blue-800",
            "border-blue-900",
            "border-indigo-100",
            "border-indigo-200",
            "border-indigo-300",
            "border-indigo-400",
            "border-indigo-500",
            "border-indigo-600",
            "border-indigo-700",
            "border-indigo-800",
            "border-indigo-900",
            "border-purple-100",
            "border-purple-200",
            "border-purple-300",
            "border-purple-400",
            "border-purple-500",
            "border-purple-600",
            "border-purple-700",
            "border-purple-800",
            "border-purple-900",
            "border-pink-100",
            "border-pink-200",
            "border-pink-300",
            "border-pink-400",
            "border-pink-500",
            "border-pink-600",
            "border-pink-700",
            "border-pink-800",
            "border-pink-900",
        ],
        "border_style": [
            "border-solid",
            "border-dashed",
            "border-dotted",
            "border-none",
            "border-double",
        ],
        "border_width": ["border", "border-0", "border-2", "border-4", "border-8"],
        "border_width-t": [
            "border-t",
            "border-t-0",
            "border-t-2",
            "border-t-4",
            "border-t-8",
        ],
        "border_width-r": [
            "border-r",
            "border-r-0",
            "border-r-2",
            "border-r-4",
            "border-r-8",
        ],
        "border_width-b": [
            "border-b",
            "border-b-0",
            "border-b-2",
            "border-b-4",
            "border-b-8",
        ],
        "border_width-l": [
            "border-l",
            "border-l-0",
            "border-l-2",
            "border-l-4",
            "border-l-8",
        ],
        "border-radius": [
            "rounded-none",
            "rounded-sm",
            "rounded",
            "rounded-lg",
            "rounded-full",
        ],
        "border-radius-t": [
            "rounded-t-none",
            "rounded-t-sm",
            "rounded-t",
            "rounded-t-lg",
            "rounded-t-full",
        ],
        "border-radius-r": [
            "rounded-r-none",
            "rounded-r-sm",
            "rounded-r",
            "rounded-r-lg",
            "rounded-r-full",
        ],
        "border-radius-b": [
            "rounded-b-none",
            "rounded-b-sm",
            "rounded-b",
            "rounded-b-lg",
            "rounded-b-full",
        ],
        "border-radius-l": [
            "rounded-l-none",
            "rounded-l-sm",
            "rounded-l",
            "rounded-l-lg",
            "rounded-l-full",
        ],
        "border-radius-tl": [
            "rounded-tl-none",
            "rounded-tl-lg",
            "rounded-tl",
            "rounded-tl-full",
            "rounded-tl-sm",
        ],
        "border-radius-tr": [
            "rounded-tr-none",
            "rounded-tr-lg",
            "rounded-tr-sm",
            "rounded-tr",
            "rounded-tr-full",
        ],
        "border-radius-br": [
            "rounded-br-none",
            "rounded-br-sm",
            "rounded-br",
            "rounded-br-lg",
            "rounded-br-full",
        ],
        "border-radius-bl": [
            "rounded-bl-none",
            "rounded-bl-sm",
            "rounded-bl",
            "rounded-bl-lg",
            "rounded-bl-full",
        ],
        "placeholder_color": [
            "placeholder-transparent",
            "placeholder-current",
            "placeholder-black",
            "placeholder-white",
            "placeholder-gray-100",
            "placeholder-gray-200",
            "placeholder-gray-300",
            "placeholder-gray-400",
            "placeholder-gray-500",
            "placeholder-gray-600",
            "placeholder-gray-700",
            "placeholder-gray-800",
            "placeholder-gray-900",
            "placeholder-red-100",
            "placeholder-red-200",
            "placeholder-red-300",
            "placeholder-red-400",
            "placeholder-red-500",
            "placeholder-red-600",
            "placeholder-red-700",
            "placeholder-red-800",
            "placeholder-red-900",
            "placeholder-orange-100",
            "placeholder-orange-200",
            "placeholder-orange-300",
            "placeholder-orange-400",
            "placeholder-orange-500",
            "placeholder-orange-600",
            "placeholder-orange-700",
            "placeholder-orange-800",
            "placeholder-orange-900",
            "placeholder-yellow-100",
            "placeholder-yellow-200",
            "placeholder-yellow-300",
            "placeholder-yellow-400",
            "placeholder-yellow-500",
            "placeholder-yellow-600",
            "placeholder-yellow-700",
            "placeholder-yellow-800",
            "placeholder-yellow-900",
            "placeholder-green-100",
            "placeholder-green-200",
            "placeholder-green-300",
            "placeholder-green-400",
            "placeholder-green-500",
            "placeholder-green-600",
            "placeholder-green-700",
            "placeholder-green-800",
            "placeholder-green-900",
            "placeholder-teal-100",
            "placeholder-teal-200",
            "placeholder-teal-300",
            "placeholder-teal-400",
            "placeholder-teal-500",
            "placeholder-teal-600",
            "placeholder-teal-700",
            "placeholder-teal-800",
            "placeholder-teal-900",
            "placeholder-blue-100",
            "placeholder-blue-200",
            "placeholder-blue-300",
            "placeholder-blue-400",
            "placeholder-blue-500",
            "placeholder-blue-600",
            "placeholder-blue-700",
            "placeholder-blue-800",
            "placeholder-blue-900",
            "placeholder-indigo-100",
            "placeholder-indigo-200",
            "placeholder-indigo-300",
            "placeholder-indigo-400",
            "placeholder-indigo-500",
            "placeholder-indigo-600",
            "placeholder-indigo-700",
            "placeholder-indigo-800",
            "placeholder-indigo-900",
            "placeholder-purple-100",
            "placeholder-purple-200",
            "placeholder-purple-300",
            "placeholder-purple-400",
            "placeholder-purple-500",
            "placeholder-purple-600",
            "placeholder-purple-700",
            "placeholder-purple-800",
            "placeholder-purple-900",
            "placeholder-pink-100",
            "placeholder-pink-200",
            "placeholder-pink-300",
            "placeholder-pink-400",
            "placeholder-pink-500",
            "placeholder-pink-600",
            "placeholder-pink-700",
            "placeholder-pink-800",
            "placeholder-pink-900",
        ],
        "flex_direction": [
            "flex-row",
            "flex-row-reverse",
            "flex-col",
            "flex-col-reverse",
        ],
        "flex_wrap": ["flex-no-wrap", "flex-wrap", "flex-wrap-reverse"],
        "align_items": [
            "items-stretch",
            "items-start",
            "items-center",
            "items-end",
            "items-baseline",
        ],
        "align_content": [
            "content-start",
            "content-center",
            "content-end",
            "content-between",
            "content-around",
        ],
        "align_self": [
            "self-auto",
            "self-start",
            "self-center",
            "self-end",
            "self-stretch",
        ],
        "justify_content": [
            "justify-start",
            "justify-center",
            "justify-end",
            "justify-between",
            "justify-around",
        ],
        "flex": ["flex-initial", "flex-1", "flex-auto", "flex-none"],
        "flex_grow": ["flex-grow", "flex-grow-0"],
        "flex_shrink": ["flex-shrink", "flex-shrink-0"],
        "order": [
            "order-first",
            "order-last",
            "order-none",
            "order-1",
            "order-2",
            "order-3",
            "order-4",
            "order-5",
            "order-6",
            "order-7",
            "order-8",
            "order-9",
            "order-10",
            "order-11",
            "order-12",
        ],
        "padding": [
            "p-0",
            "p-1",
            "p-2",
            "p-3",
            "p-4",
            "p-5",
            "p-6",
            "p-8",
            "p-10",
            "p-12",
            "p-16",
            "p-20",
            "p-24",
            "p-32",
            "p-40",
            "p-48",
            "p-56",
            "p-64",
            "p-px",
        ],
        "padding-y": [
            "py-0",
            "py-1",
            "py-2",
            "py-3",
            "py-4",
            "py-5",
            "py-6",
            "py-8",
            "py-10",
            "py-12",
            "py-16",
            "py-20",
            "py-24",
            "py-32",
            "py-40",
            "py-48",
            "py-56",
            "py-64",
            "py-px",
        ],
        "padding-x": [
            "px-0",
            "px-1",
            "px-2",
            "px-3",
            "px-4",
            "px-5",
            "px-6",
            "px-8",
            "px-10",
            "px-12",
            "px-16",
            "px-20",
            "px-24",
            "px-32",
            "px-40",
            "px-48",
            "px-56",
            "px-64",
            "px-px",
        ],
        "padding-t": [
            "pt-0",
            "pt-1",
            "pt-2",
            "pt-3",
            "pt-4",
            "pt-5",
            "pt-6",
            "pt-8",
            "pt-10",
            "pt-12",
            "pt-16",
            "pt-20",
            "pt-24",
            "pt-32",
            "pt-40",
            "pt-48",
            "pt-56",
            "pt-64",
            "pt-px",
        ],
        "padding-r": [
            "pr-0",
            "pr-1",
            "pr-2",
            "pr-3",
            "pr-4",
            "pr-5",
            "pr-6",
            "pr-8",
            "pr-10",
            "pr-12",
            "pr-16",
            "pr-20",
            "pr-24",
            "pr-32",
            "pr-40",
            "pr-48",
            "pr-56",
            "pr-64",
            "pr-px",
        ],
        "padding-b": [
            "pb-0",
            "pb-1",
            "pb-2",
            "pb-3",
            "pb-4",
            "pb-5",
            "pb-6",
            "pb-8",
            "pb-10",
            "pb-12",
            "pb-16",
            "pb-20",
            "pb-24",
            "pb-32",
            "pb-40",
            "pb-48",
            "pb-56",
            "pb-64",
            "pb-px",
        ],
        "padding-l": [
            "pl-0",
            "pl-1",
            "pl-2",
            "pl-3",
            "pl-4",
            "pl-5",
            "pl-6",
            "pl-8",
            "pl-10",
            "pl-12",
            "pl-16",
            "pl-20",
            "pl-24",
            "pl-32",
            "pl-40",
            "pl-48",
            "pl-56",
            "pl-64",
            "pl-px",
        ],
        "margin": [
            "m-0",
            "m-1",
            "m-2",
            "m-3",
            "m-4",
            "m-5",
            "m-6",
            "m-8",
            "m-10",
            "m-12",
            "m-16",
            "m-20",
            "m-24",
            "m-32",
            "m-40",
            "m-48",
            "m-56",
            "m-64",
            "m-auto",
            "m-px",
            "-m-1",
            "-m-2",
            "-m-3",
            "-m-4",
            "-m-5",
            "-m-6",
            "-m-8",
            "-m-10",
            "-m-12",
            "-m-16",
            "-m-20",
            "-m-24",
            "-m-32",
            "-m-40",
            "-m-48",
            "-m-56",
            "-m-64",
            "-m-px",
        ],
        "margin-y": [
            "my-0",
            "my-1",
            "my-2",
            "my-3",
            "my-4",
            "my-5",
            "my-6",
            "my-8",
            "my-10",
            "my-12",
            "my-16",
            "my-20",
            "my-24",
            "my-32",
            "my-40",
            "my-48",
            "my-56",
            "my-64",
            "my-auto",
            "my-px",
            "-my-1",
            "-my-2",
            "-my-3",
            "-my-4",
            "-my-5",
            "-my-6",
            "-my-8",
            "-my-10",
            "-my-12",
            "-my-16",
            "-my-20",
            "-my-24",
            "-my-32",
            "-my-40",
            "-my-48",
            "-my-56",
            "-my-64",
            "-my-px",
        ],
        "margin-x": [
            "mx-0",
            "mx-1",
            "mx-2",
            "mx-3",
            "mx-4",
            "mx-5",
            "mx-6",
            "mx-8",
            "mx-10",
            "mx-12",
            "mx-16",
            "mx-20",
            "mx-24",
            "mx-32",
            "mx-40",
            "mx-48",
            "mx-56",
            "mx-64",
            "mx-auto",
            "mx-px",
            "-mx-1",
            "-mx-2",
            "-mx-3",
            "-mx-4",
            "-mx-5",
            "-mx-6",
            "-mx-8",
            "-mx-10",
            "-mx-12",
            "-mx-16",
            "-mx-20",
            "-mx-24",
            "-mx-32",
            "-mx-40",
            "-mx-48",
            "-mx-56",
            "-mx-64",
            "-mx-px",
        ],
        "margin-t": [
            "mt-0",
            "mt-1",
            "mt-2",
            "mt-3",
            "mt-4",
            "mt-5",
            "mt-6",
            "mt-8",
            "mt-10",
            "mt-12",
            "mt-16",
            "mt-20",
            "mt-24",
            "mt-32",
            "mt-40",
            "mt-48",
            "mt-56",
            "mt-64",
            "mt-auto",
            "mt-px",
            "-mt-1",
            "-mt-2",
            "-mt-3",
            "-mt-4",
            "-mt-5",
            "-mt-6",
            "-mt-8",
            "-mt-10",
            "-mt-12",
            "-mt-16",
            "-mt-20",
            "-mt-24",
            "-mt-32",
            "-mt-40",
            "-mt-48",
            "-mt-56",
            "-mt-64",
            "-mt-px",
        ],
        "margin-r": [
            "mr-0",
            "mr-1",
            "mr-2",
            "mr-3",
            "mr-4",
            "mr-5",
            "mr-6",
            "mr-8",
            "mr-10",
            "mr-12",
            "mr-16",
            "mr-20",
            "mr-24",
            "mr-32",
            "mr-40",
            "mr-48",
            "mr-56",
            "mr-64",
            "mr-auto",
            "mr-px",
            "-mr-1",
            "-mr-2",
            "-mr-3",
            "-mr-4",
            "-mr-5",
            "-mr-6",
            "-mr-8",
            "-mr-10",
            "-mr-12",
            "-mr-16",
            "-mr-20",
            "-mr-24",
            "-mr-32",
            "-mr-40",
            "-mr-48",
            "-mr-56",
            "-mr-64",
            "-mr-px",
        ],
        "margin-b": [
            "mb-0",
            "mb-1",
            "mb-2",
            "mb-3",
            "mb-4",
            "mb-5",
            "mb-6",
            "mb-8",
            "mb-10",
            "mb-12",
            "mb-16",
            "mb-20",
            "mb-24",
            "mb-32",
            "mb-40",
            "mb-48",
            "mb-56",
            "mb-64",
            "mb-auto",
            "mb-px",
            "-mb-1",
            "-mb-2",
            "-mb-3",
            "-mb-4",
            "-mb-5",
            "-mb-6",
            "-mb-8",
            "-mb-10",
            "-mb-12",
            "-mb-16",
            "-mb-20",
            "-mb-24",
            "-mb-32",
            "-mb-40",
            "-mb-48",
            "-mb-56",
            "-mb-64",
            "-mb-px",
        ],
        "margin-l": [
            "ml-0",
            "ml-1",
            "ml-2",
            "ml-3",
            "ml-4",
            "ml-5",
            "ml-6",
            "ml-8",
            "ml-10",
            "ml-12",
            "ml-16",
            "ml-20",
            "ml-24",
            "ml-32",
            "ml-40",
            "ml-48",
            "ml-56",
            "ml-64",
            "ml-auto",
            "ml-px",
            "-ml-1",
            "-ml-2",
            "-ml-3",
            "-ml-4",
            "-ml-5",
            "-ml-6",
            "-ml-8",
            "-ml-10",
            "-ml-12",
            "-ml-16",
            "-ml-20",
            "-ml-24",
            "-ml-32",
            "-ml-40",
            "-ml-48",
            "-ml-56",
            "-ml-64",
            "-ml-px",
        ],
        "width": [
            "w-0",
            "w-1",
            "w-2",
            "w-3",
            "w-4",
            "w-5",
            "w-6",
            "w-8",
            "w-10",
            "w-12",
            "w-16",
            "w-20",
            "w-24",
            "w-32",
            "w-40",
            "w-48",
            "w-56",
            "w-64",
            "w-auto",
            "w-px",
            "w-1/2",
            "w-1/3",
            "w-2/3",
            "w-1/4",
            "w-2/4",
            "w-3/4",
            "w-1/5",
            "w-2/5",
            "w-3/5",
            "w-4/5",
            "w-1/6",
            "w-2/6",
            "w-3/6",
            "w-4/6",
            "w-5/6",
            "w-1/12",
            "w-2/12",
            "w-3/12",
            "w-4/12",
            "w-5/12",
            "w-6/12",
            "w-7/12",
            "w-8/12",
            "w-9/12",
            "w-10/12",
            "w-11/12",
            "w-full",
            "w-screen",
        ],
        "min_width": ["min-w-0", "min-w-full"],
        "max_width": [
            "max-w-xs",
            "max-w-sm",
            "max-w-md",
            "max-w-lg",
            "max-w-xl",
            "max-w-2xl",
            "max-w-3xl",
            "max-w-4xl",
            "max-w-5xl",
            "max-w-6xl",
            "max-w-full",
            "max-w-screen-sm",
            "max-w-screen-md",
            "max-w-screen-lg",
            "max-w-screen-xl",
            "max-w-none",
        ],
        "height": [
            "h-0",
            "h-1",
            "h-2",
            "h-3",
            "h-4",
            "h-5",
            "h-6",
            "h-8",
            "h-10",
            "h-12",
            "h-16",
            "h-20",
            "h-24",
            "h-32",
            "h-40",
            "h-48",
            "h-56",
            "h-64",
            "h-auto",
            "h-px",
            "h-full",
            "h-screen",
        ],
        "min_height": ["min-h-0", "min-h-full", "min-h-screen"],
        "max_height": ["max-h-full", "max-h-screen"],
        "border_collapse": ["border-collapse", "border-separate"],
        "table_layout": ["table-auto", "table-fixed"],
        "box_shadow": [
            "shadow-xs",
            "shadow-sm",
            "shadow",
            "shadow-md",
            "shadow-lg",
            "shadow-xl",
            "shadow-2xl",
            "shadow-inner",
            "shadow-outline",
            "shadow-none",
        ],
        "opacity": [
            "opacity-100",
            "opacity-75",
            "opacity-50",
            "opacity-25",
            "opacity-0",
        ],
        "appearance": ["appearance-none"],
        "cursor": [
            "cursor-auto",
            "cursor-default",
            "cursor-pointer",
            "cursor-wait",
            "cursor-text",
            "cursor-move",
            "cursor-not-allowed",
        ],
        "outline": ["outline-none"],
        "pointer_events": ["pointer-events-none", "pointer-events-auto"],
        "resize": ["resize-none", "resize", "resize-y", "resize-x"],
        "user_select": ["select-none", "select-text", "select-all", "select-auto"],
        "fill": ["fill-current"],
        "stroke": ["stroke-current"],
        "transition": [
            "transition-none",
            "transition-all",
            "transition",
            "transition-colors",
            "transition-opacity",
            "transition-shadow",
            "transition-transform",
        ],
        "duration": [
            "duration-75",
            "duration-100",
            "duration-150",
            "duration-200",
            "duration-300",
            "duration-500",
            "duration-700",
            "duration-1000",
        ],
        "transition_timing": ["ease-linear", "ease-in", "ease-out", "ease-in-out"],
        "transition_delay": [
            "delay-75",
            "delay-100",
            "delay-150",
            "delay-200",
            "delay-300",
            "delay-500",
            "delay-700",
            "delay-1000",
        ],
        "scale": [
            "scale-0",
            "scale-50",
            "scale-75",
            "scale-90",
            "scale-95",
            "scale-100",
            "scale-105",
            "scale-110",
            "scale-125",
            "scale-150",
            "scale-y-0",
            "scale-y-50",
            "scale-y-75",
            "scale-y-90",
            "scale-y-95",
            "scale-y-100",
            "scale-y-105",
            "scale-y-110",
            "scale-y-125",
            "scale-y-150",
        ],
        "rotate": [
            "rotate-0",
            "rotate-45",
            "rotate-90",
            "rotate-180",
            "-rotate-180",
            "-rotate-90",
            "-rotate-45",
        ],
        "skew": [
            "skew-x-0",
            "skew-x-3",
            "skew-x-6",
            "skew-x-12",
            "-skew-x-12",
            "-skew-x-6",
            "-skew-x-3",
            "skew-y-0",
            "skew-y-3",
            "skew-y-6",
            "skew-y-12",
            "-skew-y-12",
            "-skew-y-6",
            "-skew-y-3",
        ],
        "transform_origin": [
            "origin-center",
            "origin-top",
            "origin-top-right",
            "origin-right",
            "origin-bottom-right",
            "origin-bottom",
            "origin-bottom-left",
            "origin-left",
            "origin-top-left",
        ],
        "translate": [
            "translate-x-0",
            "translate-x-1",
            "translate-x-2",
            "translate-x-3",
            "translate-x-4",
            "translate-x-5",
            "translate-x-6",
            "translate-x-8",
            "translate-x-10",
            "translate-x-12",
            "translate-x-16",
            "translate-x-20",
            "translate-x-24",
            "translate-x-32",
            "translate-x-40",
            "translate-x-48",
            "translate-x-56",
            "translate-x-64",
            "translate-x-px",
            "-translate-x-1",
            "-translate-x-2",
            "-translate-x-3",
            "-translate-x-4",
            "-translate-x-5",
            "-translate-x-6",
            "-translate-x-8",
            "-translate-x-10",
            "-translate-x-12",
            "-translate-x-16",
            "-translate-x-20",
            "-translate-x-24",
            "-translate-x-32",
            "-translate-x-40",
            "-translate-x-48",
            "-translate-x-56",
            "-translate-x-64",
            "-translate-x-px",
            "-translate-x-full",
            "-translate-x-1/2",
            "translate-x-1/2",
            "translate-x-full",
            "translate-y-0",
            "translate-y-1",
            "translate-y-2",
            "translate-y-3",
            "translate-y-4",
            "translate-y-5",
            "translate-y-6",
            "translate-y-8",
            "translate-y-10",
            "translate-y-12",
            "translate-y-16",
            "translate-y-20",
            "translate-y-24",
            "translate-y-32",
            "translate-y-40",
            "translate-y-48",
            "translate-y-56",
            "translate-y-64",
            "translate-y-px",
            "-translate-y-1",
            "-translate-y-2",
            "-translate-y-3",
            "-translate-y-4",
            "-translate-y-5",
            "-translate-y-6",
            "-translate-y-8",
            "-translate-y-10",
            "-translate-y-12",
            "-translate-y-16",
            "-translate-y-20",
            "-translate-y-24",
            "-translate-y-32",
            "-translate-y-40",
            "-translate-y-48",
            "-translate-y-56",
            "-translate-y-64",
            "-translate-y-px",
            "-translate-y-full",
            "-translate-y-1/2",
            "translate-y-1/2",
            "translate-y-full",
        ],
        "divide_width": [
            "divide-x-0",
            "divide-x-2",
            "divide-x-4",
            "divide-x-8",
            "divide-x",
            "divide-y-0",
            "divide-y-2",
            "divide-y-4",
            "divide-y-8",
            "divide-y",
            "divide-x-reverse",
            "divide-y-reverse",
        ],
        "divide_color": [
            "divide-transparent",
            "divide-current",
            "divide-black",
            "divide-white",
            "divide-gray-100",
            "divide-gray-200",
            "divide-gray-300",
            "divide-gray-400",
            "divide-gray-500",
            "divide-gray-600",
            "divide-gray-700",
            "divide-gray-800",
            "divide-gray-900",
            "divide-red-100",
            "divide-red-200",
            "divide-red-300",
            "divide-red-400",
            "divide-red-500",
            "divide-red-600",
            "divide-red-700",
            "divide-red-800",
            "divide-red-900",
            "divide-orange-100",
            "divide-orange-200",
            "divide-orange-300",
            "divide-orange-400",
            "divide-orange-500",
            "divide-orange-600",
            "divide-orange-700",
            "divide-orange-800",
            "divide-orange-900",
            "divide-yellow-100",
            "divide-yellow-200",
            "divide-yellow-300",
            "divide-yellow-400",
            "divide-yellow-500",
            "divide-yellow-600",
            "divide-yellow-700",
            "divide-yellow-800",
            "divide-yellow-900",
            "divide-green-100",
            "divide-green-200",
            "divide-green-300",
            "divide-green-400",
            "divide-green-500",
            "divide-green-600",
            "divide-green-700",
            "divide-green-800",
            "divide-green-900",
            "divide-teal-100",
            "divide-teal-200",
            "divide-teal-300",
            "divide-teal-400",
            "divide-teal-500",
            "divide-teal-600",
            "divide-teal-700",
            "divide-teal-800",
            "divide-teal-900",
            "divide-blue-100",
            "divide-blue-200",
            "divide-blue-300",
            "divide-blue-400",
            "divide-blue-500",
            "divide-blue-600",
            "divide-blue-700",
            "divide-blue-800",
            "divide-blue-900",
            "divide-indigo-100",
            "divide-indigo-200",
            "divide-indigo-300",
            "divide-indigo-400",
            "divide-indigo-500",
            "divide-indigo-600",
            "divide-indigo-700",
            "divide-indigo-800",
            "divide-indigo-900",
            "divide-purple-100",
            "divide-purple-200",
            "divide-purple-300",
            "divide-purple-400",
            "divide-purple-500",
            "divide-purple-600",
            "divide-purple-700",
            "divide-purple-800",
            "divide-purple-900",
            "divide-pink-100",
            "divide-pink-200",
            "divide-pink-300",
            "divide-pink-400",
            "divide-pink-500",
            "divide-pink-600",
            "divide-pink-700",
            "divide-pink-800",
            "divide-pink-900",
        ],
        "divide_opacity": [
            "divide-opacity-0",
            "divide-opacity-25",
            "divide-opacity-50",
            "divide-opacity-75",
            "divide-opacity-100",
        ],
        "grid_template_cols": [
            "grid-cols-1",
            "grid-cols-2",
            "grid-cols-3",
            "grid-cols-4",
            "grid-cols-5",
            "grid-cols-6",
            "grid-cols-7",
            "grid-cols-8",
            "grid-cols-9",
            "grid-cols-10",
            "grid-cols-11",
            "grid-cols-12",
        ],
        "grid_col_start_end": [
            "col-auto",
            "col-span-1",
            "col-span-2",
            "col-span-3",
            "col-span-4",
            "col-span-5",
            "col-span-6",
            "col-span-7",
            "col-span-8",
            "col-span-9",
            "col-span-10",
            "col-span-11",
        ],
        "grid_template_rows": [
            "grid-rows-1",
            "grid-rows-2",
            "grid-rows-3",
            "grid-rows-4",
            "grid-rows-5",
            "grid-rows-6",
            "grid-rows-none",
        ],
        "grid_row_start_end": [
            "row-auto",
            "row-span-1",
            "row-span-2",
            "row-span-3",
            "row-span-4",
            "row-span-5",
            "row-span-6",
            "row-start-1",
            "row-start-2",
            "row-start-3",
            "row-start-4",
            "row-start-5",
            "row-start-6",
            "row-start-7",
            "row-start-auto",
            "row-end-1",
            "row-end-2",
            "row-end-3",
            "row-end-4",
            "row-end-5",
            "row-end-6",
            "row-end-7",
            "row-end-auto",
        ],
        "grid_gap": [
            "gap-0",
            "gap-1",
            "gap-2",
            "gap-3",
            "gap-4",
            "gap-5",
            "gap-6",
            "gap-8",
            "gap-10",
            "gap-12",
            "gap-16",
            "gap-20",
            "gap-24",
            "gap-32",
            "gap-40",
            "gap-48",
            "gap-56",
            "gap-64",
            "gap-px",
            "row-gap-0",
            "row-gap-1",
            "row-gap-2",
            "row-gap-3",
            "row-gap-4",
            "row-gap-5",
            "row-gap-6",
            "row-gap-8",
            "row-gap-10",
            "row-gap-12",
            "row-gap-16",
            "row-gap-20",
            "row-gap-24",
            "row-gap-32",
            "row-gap-40",
            "row-gap-48",
            "row-gap-56",
            "row-gap-64",
            "row-gap-px",
            "col-gap-0",
            "col-gap-1",
            "col-gap-2",
            "col-gap-3",
            "col-gap-4",
            "col-gap-5",
            "col-gap-6",
            "col-gap-8",
            "col-gap-10",
            "col-gap-12",
            "col-gap-16",
            "col-gap-20",
            "col-gap-24",
            "col-gap-32",
            "col-gap-40",
            "col-gap-48",
            "col-gap-56",
            "col-gap-64",
            "col-gap-px",
        ],
        "grid_auto_flow": [
            "grid-flow-row",
            "grid-flow-col",
            "grid-flow-row-dense",
            "grid-flow-col-dense",
        ],
        "space_between": [
            "space-x-0",
            "space-x-1",
            "space-x-2",
            "space-x-3",
            "space-x-4",
            "space-x-5",
            "space-x-6",
            "space-x-8",
            "space-x-10",
            "space-x-12",
            "space-x-16",
            "space-x-20",
            "space-x-24",
            "space-x-32",
            "space-x-40",
            "space-x-48",
            "space-x-56",
            "space-x-64",
            "space-x-px",
            "-space-x-1",
            "-space-x-2",
            "-space-x-3",
            "-space-x-4",
            "-space-x-5",
            "-space-x-6",
            "-space-x-8",
            "-space-x-10",
            "-space-x-12",
            "-space-x-16",
            "-space-x-20",
            "-space-x-24",
            "-space-x-32",
            "-space-x-40",
            "-space-x-48",
            "-space-x-56",
            "-space-x-64",
            "-space-x-px",
            "space-y-0",
            "space-y-1",
            "space-y-2",
            "space-y-3",
            "space-y-4",
            "space-y-5",
            "space-y-6",
            "space-y-8",
            "space-y-10",
            "space-y-12",
            "space-y-16",
            "space-y-20",
            "space-y-24",
            "space-y-32",
            "space-y-40",
            "space-y-48",
            "space-y-56",
            "space-y-64",
            "space-y-px",
            "-space-y-1",
            "-space-y-2",
            "-space-y-3",
            "-space-y-4",
            "-space-y-5",
            "-space-y-6",
            "-space-y-8",
            "-space-y-10",
            "-space-y-12",
            "-space-y-16",
            "-space-y-20",
            "-space-y-24",
            "-space-y-32",
            "-space-y-40",
            "-space-y-48",
            "-space-y-56",
            "-space-y-64",
            "-space-y-px",
            "space-x-reverse",
            "space-y-reverse",
        ],
        "animation": [
            "animate-none",
            "animate-spin",
            "animate-ping",
            "animate-pulse",
            "animate-bounce",
        ],
    }

    @staticmethod
    def create_reverse_dict(tw):
        d = {}
        for k, v in tw.items():
            for j in v:
                d[j] = k
        return d

    tw_reverse_dict = create_reverse_dict.__func__(tw_dict)

    def set_class(self, tw_class, modifier=""):
        if modifier and modifier not in Tailwind.pseudo_classes:
            raise Exception(f"No Tailwind pseudo-class (modifier) named {modifier}")
        if tw_class not in Tailwind.tw_reverse_dict:
            raise Exception(f"No Tailwind class named {tw_class}")
        class_list = self.classes.split()
        if not modifier:
            for i in class_list:
                if i in Tailwind.tw_dict[Tailwind.tw_reverse_dict[tw_class]]:
                    class_list.remove(i)
            class_list.append(tw_class)
        else:
            tw_dict_modified = [
                f"{modifier}:" + i
                for i in Tailwind.tw_dict[Tailwind.tw_reverse_dict[tw_class]]
            ]
            tw_class_modified = f"{modifier}:{tw_class}"
            for i in class_list:
                if i in tw_dict_modified:
                    class_list.remove(i)
            class_list.append(tw_class_modified)
        self.classes = " ".join(class_list)
        return self.classes

    def set_classes(self, class_list):
        # Takes a string of tailwind classes and sets them all
        for c in class_list.split():
            c = c.split(":")
            if len(c) > 1:
                self.set_class(c[1], c[0])
            else:
                self.set_class(c[0])



"""
Created on 2022-08-30

@author: @sandeep-gh
@author: wf

see https://github.com/justpy-org/justpy/issues/478 
and https://github.com/justpy-org/justpy/issues/389
"""
from tests.base_client_test import BaseClienttest
import justpy as jp
from starlette.responses import PlainTextResponse
from jpcore.justpy_config import HTML_404_PAGE

@jp.app.route("/plaintext")
async def plainText(_request):
    return PlainTextResponse("Plaintext")

@jp.app.route("/urlFor/{name}")
async def urlFor(request):
    name=request.path_params["name"]
    print(f"calling url_for {name}")
    url=request.url_for(name)
    return PlainTextResponse(url)

@jp.app.route("/greet/{name}")
@jp.app.response
def greeting_function(request):
    wp = jp.WebPage()
    name = f"""{request.path_params["name"]}"""
    wp.add(jp.P(text=f"Hello there, {name}!", classes="text-5xl m-2"))
    return wp

@jp.app.jproute("/bye")
def bye_function(_request):
    wp = jp.WebPage()
    wp.add(jp.P(text="Bye bye!", classes="text-5xl m-2"))
    return wp

@jp.app.route("/hello",name="hello")
@jp.app.response
def hello_function(_request):
    wp = jp.WebPage()
    wp.add(jp.P(text="Hello there!", classes="text-5xl m-2"))
    # print("request  = ", request.url_for("hello"))
    return wp

class TestRouteAndUrlFor(BaseClienttest):
    """
    test the url_for functionality
    """

    def setUp(self, debug=False, profile=True):
        BaseClienttest.setUp(self, debug=debug, profile=profile)
        #self.app.prioritize_routes()
        
    def testStarletteRoutes(self):
        """
        test the starlette Routes
        """
        debug=True
        for i,route in enumerate(self.app.router.routes):
            route_as_text = self.app.route_as_text(route)
            if debug:
                print(f"{i}:{route_as_text}")
            

    def testJpRoute(self):
        """
        Test the routing
        """
        routes=self.app.router.routes
        self.assertTrue(len(routes)>=9)
        debug = self.debug
        debug=True
        for path in ["/bye", "/hello", "/greet/{name}"]:
            found=False
            for route in routes:
                found=route.path==path
                if found:
                    break
            self.assertTrue(found)

    def testUrlFor(self):
        """
        Test url for functionality
        """
        for name,url in [
            ("plainText","plaintext"),
            ("hello","hello"),
            ("bye_function","bye")
        ]:
            path=f"/urlFor/{name}"
            response=self.checkResponse(path)
            urlfor=response.text
            print(urlfor)
            self.assertEqual(urlfor,f"http://testserver/{url}")
            pass
    
    def testResponses(self):
        
        for path in [
            "/plaintext",
            #"/hello","/greet/eli","/bye"
            ]:
            _response=self.checkResponse(path,debug=True)

    def testInvalidPath(self):
        """
        test handling an invalid path
        """
        response = self.checkResponse("/invalidpath", 404)
        #self.assertEqual(HTML_404_PAGE, response.text)

    def testStaticPath(self):
        """
        test handling static content
        """
        response = self.checkResponse("/templates/js/justpy_core.js")
        self.assertTrue("class JustpyCore" in response.text)
        response = self.checkResponse("/templates/css/invalid_css.css", 404)

    def testHtmlContent(self):
        # see https://www.starlette.io/testclient/
        response = self.checkResponse("/hello")
        debug = self.debug
        debug = True
        lines = response.text.split("\n")
        if debug:
            print(response.text)
            print(f"{len(lines)} lines")
        self.assertTrue(response.text.startswith("<!DOCTYPE html>"))
        self.assertTrue(len(lines) < 80)



# Justpy Tutorial demo input_demo1 from docs/tutorial/input.md
import justpy as jp

input_classes = "m-2 bg-gray-200 border-2 border-gray-200 rounded w-64 py-2 px-4 text-gray-700 focus:outline-none focus:bg-white focus:border-purple-500"

def input_demo1(request):
    wp = jp.WebPage()
    in1 = jp.Input(a=wp, classes=input_classes, placeholder='Please type here')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("input_demo1",input_demo1)



import hjson

from .htmlcomponents import *
from .htmlcomponents import _tag_class_dict, parse_dict
from addict import Dict

quasar_directives = [
    "v-close-popup",
    "v-close-menu",
    "v-ripple",
    "v-model",
    "v-close-dialog",
]


class QuasarPage(WebPage):
    """
    a Quasar based WebPage
    """

    def __init__(self, **kwargs):
        """
        constructor
        """
        super().__init__(**kwargs)
        self.tailwind = kwargs.get("tailwind", False)
        self.template_file = "quasar.html"
        self.quasar = True

    async def set_dark_mode(self, flag):
        try:
            websocket_dict = WebPage.sockets[self.page_id]
        except:
            return self
        for websocket in list(websocket_dict.values()):
            try:
                WebPage.loop.create_task(
                    websocket.send_json({"type": "page_mode_update", "dark": flag})
                )
            except:
                print("Problem with websocket in page update, ignoring")
        return self

    async def toggle_full_screen(self):
        await self.run_javascript("Quasar.AppFullscreen.toggle()")
        return True

class QDiv(Div):
    slots = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.vue_type = "quasar_component"
        self.directives = quasar_directives

    def __setattr__(self, key, value):
        if key in self.slots:
            q_slot = key[: key.index("_slot")].replace("_", "-")
            self.add_scoped_slot(q_slot, value)
        else:
            super().__setattr__(key, value)


class _QInputBase(Input):
    """
    Base Class for Quasar Input
    """

    slots = []

    def __init__(self, **kwargs):
        """
        constructor
        """
        self.disable_input_event = False
        super().__init__(**kwargs)
        self.vue_type = "quasar_component"
        self.directives = quasar_directives
        self.disable_events = False
        self.prop_list = []
        self.attributes = []
        self.evaluate_prop = []

    def __setattr__(self, key, value):
        if key == "options":
            if isinstance(value, str):
                self.load_json(value)
            else:
                self.__dict__[key] = value
        elif key in self.slots:
            q_slot = key[: key.index("_slot")].replace("_", "-")
            self.add_scoped_slot(q_slot, value)
        else:
            self.__dict__[key] = value

    def model_update(self):
        update_value = self.model[0].data[self.model[1]]
        self.value = update_value

    def load_json(self, options_string):
        self.options = hjson.loads(options_string.encode("ascii", "ignore"))
        return self.options

    def load_json_from_file(self, file_name):
        with open(file_name, "r") as f:
            self.options = hjson.loads(f.read().encode("ascii", "ignore"))
        return self.options

    def convert_object_to_dict(self):

        d = super().convert_object_to_dict()
        if self.disable_events:
            d["events"] = []
        d["evaluate_prop"] = self.evaluate_prop
        d["disable_input_event"] = self.disable_input_event
        return d


QInputBase = _QInputBase


@parse_dict
class QInput(_QInputBase):
    """
    https://quasar.dev/vue-components/input
    """

    html_tag = "q-input"
    slots = [
        "default_slot",
        "before_slot",
        "after_slot",
        "error_slot",
        "hint_slot",
        "counter_slot",
        "loading_slot",
        "prepend_slot",
        "append_slot",
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "mask",
            "fill-mask",
            "unmasked-value",
            "error",
            "error-message",
            "rules",
            "lazy-rules",
            "label",
            "stack-label",
            "hint",
            "hide-hint",
            "prefix",
            "suffix",
            "color",
            "bg-color",
            "dark",
            "filled",
            "outlined",
            "borderless",
            "standout",
            "bottom-slots",
            "rounded",
            "square",
            "dense",
            "items-aligned",
            "disable",
            "readonly",
            "value",
            "type",
            "debounce",
            "counter",
            "maxlength",
            "autogrow",
            "autofocus",
            "input-class",
            "input-style",
            "clearable",
            "clear-icon",
            "placeholder",
            "min",
            "max",
            "loading",
        ]

        self.allowed_events = [
            "keypress",
            "input",
            "focus",
            "blur",
            "change",
        ]  # Not different from focus and blur in documentation
        self.set_keyword_events(**kwargs)
        self.evaluate_prop = ["rules"]


class QInputBlur(QInput):
    def before_event_handler(self, msg):
        logging.debug(
            "%s %s %s %s %s", "before ", self.type, msg.event_type, msg.input_type, msg
        )
        if hasattr(self, "model"):
            self.model[0].data[self.model[1]] = msg.value
        self.value = msg.value

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["disable_input_event"] = True
        if "blur" not in self.events:
            self.events.append("blur")
        return d


class QInputChange(QInput):
    def before_event_handler(self, msg):
        logging.debug(
            "%s %s %s %s %s", "before ", self.type, msg.event_type, msg.input_type, msg
        )
        if hasattr(self, "model"):
            self.model[0].data[self.model[1]] = msg.value
        self.value = msg.value

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["disable_input_event"] = True
        try:
            d["events"].remove("input")
        except:
            pass
        if "change" not in self.events:
            self.events.append("change")
        return d


@parse_dict
class QField(QDiv):
    html_tag = "q-field"
    slots = [
        "default_slot",
        "before_slot",
        "prepend_slot",
        "append_slot",
        "after_slot",
        "label_slot",
        "error_slot",
        "hint_slot",
        "counter_slot",
        "loading_slot",
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "error",
            "rules",
            "reactive-rules",
            "lazy-rules",
            "autofocus",
            "name",
            "error-message",
            "no-error-icon",
            "label",
            "stack-label",
            "hint",
            "hide-hint",
            "prefix",
            "suffix",
            "loading",
            "clearable",
            "clear-icon",
            "label-slot",
            "bottom-slots",
            "counter",
            "maxlength",
            "disable",
            "readonly",
            "label-color",
            "color",
            "bg-color",
            "dark",
            "filled",
            "outlined",
            "borderless",
            "standout",
            "hide-bottom-space",
            "rounded",
            "square",
            "dense",
            "item-aligned",
            "value",
        ]

        self.allowed_events = ["clear", "input", "focus", "blur"]


@parse_dict
class QSelect(_QInputBase):
    html_tag = "q-select"
    slots = [
        "default_slot",
        "before_slot",
        "after_slot",
        "error_slot",
        "hint_slot",
        "counter_slot",
        "loading_slot",
        "selected_slot",
        "nooption_slot",
        "selecteditem_slot",
        "option_slot",
        "prepend_slot",
        "append_slot",
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = "object"
        self.value = kwargs.get("value", None)
        self.prop_list = [
            "color",
            "bg-color",
            "dark",
            "filled",
            "outlined",
            "borderless",
            "standout",
            "hide-bottom-space",
            "rounded",
            "square",
            "dense",
            "popup-content-class",
            "popup-content-style",
            "error",
            "rules",
            "lazy-rules",
            "loading",
            "clearable",
            "autofocus",
            "hide-dropdown-icon",
            "fill-input",
            "new-value-mode",
            "transition-show",
            "transition-hide",
            "error-message",
            "no-error-icon",
            "label",
            "stack-label",
            "hint",
            "hide-hint",
            "prefix",
            "suffix",
            "clear-icon",
            "bottom-slots",
            "counter",
            "items-aligned",
            "dropdown-icon",
            "use-input",
            "input-debounce",
            "value",
            "multiple",
            "emit-value",
            "options",
            "options-value",
            "option-label",
            "option-disable",
            "options-dense",
            "options-dark",
            "options-selected-class",
            "options-cover",
            "options-sanitize",
            "map-options",
            "display-value",
            "display-value-sanitize",
            "hide-selected",
            "max-values",
            "use-chips",
            "disable",
            "readonly",
            "behavior",
            "input-class",
            "input-style",
            "virtual-scroll-slice-size",
            "virtual-scroll-item-size",
            "virtual-scroll-sticky-size-start",
            "virtual-scroll-sticky-size-end",
        ]

        self.allowed_events = [
            "input",
            "remove",
            "add",
            "new-value",
            "filter",
            "filter-abort",
            "focus",
            "blur",
            "clear",
        ]
        # self.set_keyword_events(**kwargs)


@parse_dict
class QOptionGroup(_QInputBase):
    html_tag = "q-option-group"
    slots = []

    def __init__(self, **kwargs):
        self.options = []
        super().__init__(**kwargs)
        # Type: radio | checkbox | toggle https://quasar.dev/vue-components/option-group
        if self.type == "checkbox" or self.type == "toggle":
            self.value = kwargs.get("value", [])
        else:
            self.type = "radio"
            self.value = kwargs.get("value", "")
        self.prop_list = [
            "keep-color",
            "type",
            "left-label",
            "inline",
            "value",
            "options",
            "disable",
            "size",
            "color",
            "dark",
            "dense",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)

    def before_event_handler(self, msg):
        logging.debug(
            "%s %s %s %s %s", "before ", self.type, msg.event_type, msg.input_type, msg
        )
        if hasattr(self, "model"):
            self.model[0].data[self.model[1]] = msg.value
        self.value = msg.value

    def convert_object_to_dict(self):

        d = super().convert_object_to_dict()
        d["events"] = ["before", "input"]
        return d


@parse_dict
class QBtnToggle(_QInputBase):
    html_tag = "q-btn-toggle"
    slots = []

    def __init__(self, **kwargs):
        self.options = []
        super().__init__(**kwargs)
        self.type = "object"
        self.value = kwargs.get("value", None)
        self.prop_list = [
            "spread",
            "no-caps",
            "no-wrap",
            "stack",
            "stretch",
            "value",
            "options",
            "readonly",
            "disable",
            "ripple",
            "color",
            "text-color",
            "toggle-color",
            "toggle-text-color",
            "outline",
            "flat",
            "unelevated",
            "rounded",
            "push",
            "glossy",
            "size",
            "dense",
            "clearable",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)


@parse_dict
class QSlider(_QInputBase):
    # TODO: Deal with label-value prop, for now use suffix
    html_tag = "q-slider"
    slots = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = "float"
        self.value = kwargs.get("value", 0)
        # The label suffix is a stop gap till full label-value prop implemented
        self.label_suffix = kwargs.get("label_suffix", "")
        self.prop_list = [
            "label-always",
            "snap",
            "label",
            "label-value",
            "label-always",
            "markers",
            "tabindex",
            "vertical",
            "value",
            "min",
            "max",
            "step",
            "disable",
            "readonly",
            "color",
            "label-color",
            "dark",
            "dense",
        ]
        self.allowed_events = ["input", "change"]
        # self.set_keyword_events(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        if self.label_suffix:
            d["attrs"]["label-value"] = str(self.value) + self.label_suffix
        return d


@parse_dict
class QRange(_QInputBase):
    # TODO: Deal with left-label-value  and right-label-value props, for now use suffix for both right and left

    slots = []
    html_tag = "q-range"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = "object"
        self.value = kwargs.get("value", {"min": 0, "max": 100})
        # The label suffix is a stop gap till full label-value prop implemented
        self.label_suffix = kwargs.get("label_suffix", "")
        self.prop_list = [
            "drag-range",
            "drag-range-only",
            "left-label-color",
            "right-label-color",
            "left-label-value",
            "right-label-value",
            "label-always",
            "snap",
            "label",
            "label-always",
            "markers",
            "tabindex",
            "value",
            "min",
            "max",
            "step",
            "disable",
            "readonly",
            "color",
            "dark",
            "dense",
        ]
        self.allowed_events = ["input", "change"]

        # self.set_keyword_events(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        if self.label_suffix:
            d["attrs"]["label-value"] = str(self.value) + self.label_suffix
        return d


@parse_dict
class QRating(_QInputBase):
    slots = []
    html_tag = "q-rating"

    def __init__(self, **kwargs):
        self.no_reset = False
        super().__init__(**kwargs)
        self.debounce = 30
        self.prop_list = [
            "icon",
            "max",
            "value",
            "no-reset",
            "readonly",
            "disable",
            "color",
            "size",
            "icon-selected",
            "icon-half",
            "no-dimming",
            "color-selected",
            "color-half",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)

    def before_event_handler(self, msg):
        if msg.event_type not in ["input"]:
            return

        if hasattr(self, "model"):
            self.model[0].data[self.model[1]] = msg.value

        if not self.no_reset and (self.value == msg.value):
            self.value = 0
        else:
            self.value = msg.value


@parse_dict
class QTime(_QInputBase):
    slots = []
    html_tag = "q-time"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Options prop not supported and left out from list. It is a function
        self.prop_list = [
            "landscape",
            "format24h",
            "options",
            "hour-options",
            "minute-options",
            "second-options",
            "with-seconds",
            "now-btn",
            "value",
            "mask",
            "locale",
            "calendar",
            "readonly",
            "disable",
            "color",
            "text-color",
            "dark",
            "square",
            "flat",
            "bordered",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)


@parse_dict
class QDate(_QInputBase):
    slots = []
    html_tag = "q-date"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Options prop not supported and left out from list. It is a function
        self.prop_list = [
            "landscape",
            "title",
            "subtitle",
            "today-btn",
            "minimal",
            "readonly",
            "disable",
            "color",
            "text-color",
            "dark",
            "event-color",
            "value",
            "mask",
            "locale",
            "calendar",
            "emit-immediately",
            "default-year-month",
            "default-view",
            "events-date",
            "options-date",
            "first-day-of-week",
            "square",
            "flat",
            "bordered",
        ]
        self.allowed_events = ["input"]
        self.evaluate_prop = []
        # self.set_keyword_events(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        try:
            d["attrs"]["events"] = self.events_date
        except:
            pass
        try:
            d["attrs"]["options"] = self.options_date
        except:
            pass
        return d


@parse_dict
class QCheckbox(_QInputBase):
    slots = ["default_slot"]
    html_tag = "q-checkbox"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = "object"
        self.value = kwargs.get("value", False)
        self.prop_list = [
            "keep-color",
            "indeterminate-value",
            "toggle-indeterminate",
            "tabindex",
            "label",
            "left-label",
            "name",
            "toggle-order",
            "value",
            "val",
            "true-value",
            "false-value",
            "disable",
            "color",
            "dark",
            "dense",
            "size",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        return d


@parse_dict
class QToggle(_QInputBase):
    slots = ["default_slot"]
    html_tag = "q-toggle"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.debounce = 50
        self.type = "object"
        self.value = kwargs.get("value", False)
        self.prop_list = [
            "keep-color",
            "icon",
            "checked-icon",
            "unchecked-icon",
            "tabindex",
            "label",
            "left-label",
            "value",
            "val",
            "true-value",
            "false-value",
            "disable",
            "color",
            "dark",
            "dense",
            "size",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        return d


@parse_dict
class QForm(QDiv):
    slots = ["default_slot"]
    html_tag = "q-form"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["autofocus", "no-error-focus", "no-reset-focus", "greedy"]
        self.allowed_events = ["submit", "before"]

        def default_submit(self, msg):
            print("Default form submit", msg.form_data)
            return True

        if not self.has_event_function("submit"):
            # If an event handler is not  assigned, the front end cannot stop the default page request that happens when a form is submitted
            self.on("submit", default_submit)


@parse_dict
class QAvatar(QDiv):
    slots = ["default_slot"]
    html_tag = "q-avatar"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "icon",
            "size",
            "font-size",
            "color",
            "text-color",
            "square",
            "rounded",
        ]


@parse_dict
class QAjaxBar(QDiv):
    slots = []
    html_tag = "q-ajax-bar"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["skip-hijack", "reverse", "position", "size", "color"]
        self.allowed_events = ["start", "stop"]
        self.set_keyword_events(**kwargs)


@parse_dict
class QBadge(QDiv):
    slots = ["default_slot"]
    html_tag = "q-badge"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "color",
            "text-color",
            "floating",
            "transparent",
            "multi-line",
            "label",
            "align",
            "outline",
        ]


@parse_dict
class QBanner(QDiv):
    slots = ["default_slot", "avatar_slot", "action_slot"]
    html_tag = "q-banner"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["inline-actions", "dense", "rounded"]


@parse_dict
class QBar(QDiv):
    slots = ["default_slot"]
    html_tag = "q-bar"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["dense", "dark"]


@parse_dict
class QToolbar(QDiv):
    slots = ["default_slot"]
    html_tag = "q-toolbar"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["inset"]


@parse_dict
class QToolbarTitle(QDiv):
    slots = ["default_slot"]
    html_tag = "q-toolbar-title"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["shrink"]


@parse_dict
class QBreadcrumbs(QDiv):
    # DOES NOT WORK
    html_tag = "q-breadcrumbs"

    def __init__(self, **kwargs):
        self.separator = "/"
        super().__init__(**kwargs)
        self.prop_list = [
            "separator",
            "active-color",
            "gutter",
            "separator-color",
            "align",
        ]
        self.separator = "h"
        self.separator_color = "primary"


@parse_dict
class QBreadcrumbsEl(QDiv):
    slots = []
    html_tag = "q-breadcrumbs-el"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "to",
            "exact",
            "append",
            "replace",
            "active-class",
            "exact-active-class",
            "label",
            "icon",
        ]


@parse_dict
class QBtn(QDiv):
    slots = ["default_slot", "loading_slot"]
    html_tag = "q-btn"

    def __init__(self, **kwargs):
        self.loading = False
        super().__init__(**kwargs)
        self.prop_list = [
            "ripple",
            "type",
            "to",
            "replace",
            "label",
            "icon",
            "icon-right",
            "round",
            "outline",
            "flat",
            "unelevated",
            "rounded",
            "push",
            "glossy",
            "size",
            "fab",
            "fab-mini",
            "color",
            "text-color",
            "no-caps",
            "no-wrap",
            "dense",
            "tabindex",
            "align",
            "stack",
            "stretch",
            "loading",
            "disable",
            "percentage",
            "dark-percentage",
            "href",
            "target",
            "download",
        ]


QButton = QBtn


@parse_dict
class QBtnGroup(QDiv):
    slots = ["default_slot"]
    html_tag = "q-btn-group"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "spread",
            "stretch",
            "outline",
            "flat",
            "unelevated",
            "rounded",
            "push",
            "glossy",
        ]


@parse_dict
class QBtnDropdown(_QInputBase):
    slots = ["default_slot", "label_slot"]
    html_tag = "q-btn-dropdown"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = "boolean"
        self.value = bool(self.value)
        self.prop_list = [
            "loading",
            "split",
            "disable-main-btn",
            "disable-dropdown",
            "persistent",
            "auto-close",
            "label",
            "icon",
            "icon-right",
            "no-caps",
            "no-wrap",
            "align",
            "stack",
            "stretch",
            "dropdown-icon",
            "type",
            "tabindex",
            "value",
            "cover",
            "menu-anchor",
            "menu-self",
            "to",
            "replace",
            "disable",
            "ripple",
            "round",
            "outline",
            "flat",
            "unelevated",
            "rounded",
            "push",
            "size",
            "fab",
            "fab-mini",
            "color",
            "text-color",
            "dense",
            "content-style",
            "content-class",
        ]
        self.allowed_events = [
            "input",
            "show",
            "before_show",
            "hide",
            "before_hide",
            "click",
        ]
        self.set_keyword_events(**kwargs)


@parse_dict
class QMenu(_QInputBase):
    slots = ["default_slot"]
    html_tag = "q-menu"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = "boolean"
        self.value = bool(self.value)
        self.prop_list = [
            "target",
            "context-menu",
            "fit",
            "no-parent-event",
            "touch-position",
            "persistent",
            "auto-close",
            "no-focus",
            "transition-show",
            "transition-hide",
            "value",
            "cover",
            "anchor",
            "self",
            "offset",
            "content-class",
            "content-style",
            "square",
            "max-height",
            "max-width",
        ]
        self.allowed_events = [
            "input",
            "show",
            "before_show",
            "hide",
            "before_hide",
            "escape_key",
        ]
        self.set_keyword_events(**kwargs)


@parse_dict
class QCard(QDiv):
    slots = ["default_slot"]
    html_tag = "q-card"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["dark", "square", "flat", "bordered"]


@parse_dict
class QCardSection(QDiv):
    slots = ["default_slot"]
    html_tag = "q-card-section"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["horizontal"]


@parse_dict
class QCardActions(QDiv):
    slots = ["default_slot"]
    html_tag = "q-card-actions"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["align", "vertical"]


@parse_dict
class QTabs(_QInputBase):
    slots = ["default_slot"]
    html_tag = "q-tabs"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.get("value", "")
        self.prop_list = [
            "breakpoint",
            "vertical",
            "align",
            "left-icon",
            "right-icon",
            "shrink",
            "switch-indicator",
            "narrow-indicator",
            "inline-label",
            "no-caps",
            "value",
            "active-color",
            "active-bg-color",
            "indicator-color",
            "dense",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)


@parse_dict
class QTab(QDiv):
    slots = ["default_slot"]
    html_tag = "q-tab"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "icon",
            "label",
            "alert",
            "no-caps",
            "name",
            "tabindex",
            "disable",
            "ripple",
        ]


@parse_dict
class QTabPanels(_QInputBase):
    # Does not work
    slots = ["default_slot"]
    html_tag = "q-tab-panels"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.get("value", "")
        self.prop_list = [
            "keep-alive",
            "animated",
            "infinite",
            "swipeable",
            "transition-prev",
            "transition-next",
            "value",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)


@parse_dict
class QTabPanel(QDiv):
    slots = ["default_slot"]
    html_tag = "q-tab-panel"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["name", "disable"]


@parse_dict
class QSplitter(_QInputBase):
    slots = ["before_slot", "after_slot", "separator_slot"]
    html_tag = "q-splitter"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.get("value", 50.0)
        self.limits = kwargs.get("limits", [0, 100])
        self.type = "float"
        self.prop_list = [
            "horizontal",
            "limits",
            "value",
            "disable",
            "before-class",
            "after-class",
            "separator-class",
            "separator-style",
            "dark",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)


@parse_dict
class QKnob(_QInputBase):
    slots = ["default_slot"]
    html_tag = "q-knob"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.get("value", 0)
        self.type = "float"
        self.prop_list = [
            "show-value",
            "angle",
            "tabindex",
            "value",
            "min",
            "max",
            "step",
            "disable",
            "readonly",
            "color",
            "center-color",
            "track-color",
            "size",
            "font-size",
            "thickness",
        ]
        self.allowed_events = ["input", "change", "drag_value"]
        self.set_keyword_events(**kwargs)


@parse_dict
class QPagination(_QInputBase):
    slots = []
    html_tag = "q-pagination"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.get("value", 1)
        self.type = "float"
        self.prop_list = [
            "input",
            "boundary-links",
            "direction-links",
            "boundary-numbers",
            "ellipses",
            "value",
            "min",
            "max",
            "max-pages",
            "disable",
            "color",
            "text-color",
            "input-style",
            "input-class",
            "size",
            "icon-prev",
            "icon-next",
            "icon-first",
            "icon-last",
            "ripple",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)


@parse_dict
class QChatMessage(QDiv):
    slots = ["default_slot", "avatar_slot"]
    html_tag = "q-chat-message"

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        self.prop_list = [
            "label-sanitize",
            "name-sanitize",
            "text-sanitize",
            "stamp-sanitize",
            "sent",
            "label",
            "name",
            "avatar",
            "text",
            "stamp",
            "bg-color",
            "text-color",
            "size",
        ]

    def convert_object_to_dict(self):

        d = super().convert_object_to_dict()
        try:
            d.pop("text")
        except:
            pass
        return d


@parse_dict
class QChip(QDiv):
    slots = ["default_slot"]
    html_tag = "q-chip"

    def __init__(self, **kwargs):
        self.selected = None
        super().__init__(**kwargs)
        self.prop_list = [
            "icon",
            "icon-right",
            "label",
            "tabindex",
            "value",
            "selected",
            "clickable",
            "removable",
            "disable",
            "ripple",
            "dense",
            "color",
            "text-color",
            "square",
            "outline",
        ]
        self.allowed_events = ["update:selected", "click", "remove"]
        self.set_keyword_events(**kwargs)

        def chip_remove(self, message):
            self.show = False
            self.value = False

        self.on("remove", chip_remove)

    @staticmethod
    def chip_select(self, message):
        self.selected = not self.selected

    def convert_object_to_dict(self):

        d = super().convert_object_to_dict()
        if self.selected is not None:
            self.on("update:selected", self.chip_select)
        else:
            self.remove_event("update:selected")
        return d


@parse_dict
class QCircularProgress(QDiv):
    slots = ["default_slot"]
    html_tag = "q-circular-progress"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "indeterminate",
            "show-value",
            "reverse",
            "angle",
            "value",
            "min",
            "max",
            "color",
            "center-color",
            "track-color",
            "size",
            "font-size",
            "thickness",
        ]


@parse_dict
class QLinearProgress(QDiv):
    slots = ["default_slot"]
    html_tag = "q-linear-progress"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "buffer",
            "reverse",
            "indeterminate",
            "query",
            "stripe",
            "value",
            "color",
            "track-color",
            "dark",
            "rounded",
        ]


@parse_dict
class QColor(_QInputBase):
    slots = []
    html_tag = "q-color"

    def __init__(self, **kwargs):
        self.value = ""
        self.format_model = "hex"
        super().__init__(**kwargs)
        self.prop_list = [
            "default-view",
            "no-header",
            "no-footer",
            "value",
            "default-value",
            "format-model",
            "disable",
            "readonly",
            "dark",
        ]
        self.allowed_events = ["input", "change"]
        # self.set_keyword_events(**kwargs)


@parse_dict
class QPopupProxy(_QInputBase):
    html_tag = "q-popup-proxy"
    slots = ["default_slot"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value = kwargs.get("value", False)
        # self.prop_list = ['target', 'context-menu', 'breakpoint', 'value']
        # self.qdialog_prop_list = ['persistent', 'no-esc-dsimiss', 'no-backdrop-dismiss', 'no-route-dismiss',
        #                           'auto-close',
        #                           'transition-show', 'transition-hide', 'no-refocus', 'no-focus', 'seamless',
        #                           'maximized',
        #                           'full-width', 'full-height', 'position', 'value', 'content-class', 'content-style',
        #                           'square']
        # self.qmenu_prop_list = ['target', 'context-menu', 'fit', 'no-parent-event', 'touch-position', 'persistent',
        #                         'auto-close', 'no-focus', 'transition-show', 'transition-hide', 'value', 'cover',
        #                         'anchor', 'self',
        #                         'offset', 'content-class', 'content-style', 'square', 'max-height', 'max-width']
        # Union of all three. QDialog and QMenu props are passed through
        self.prop_list = [
            "touch-position",
            "context-menu",
            "max-width",
            "no-route-dismiss",
            "content-style",
            "full-width",
            "max-height",
            "no-esc-dsimiss",
            "cover",
            "no-parent-event",
            "square",
            "no-backdrop-dismiss",
            "value",
            "transition-hide",
            "breakpoint",
            "content-class",
            "seamless",
            "transition-show",
            "fit",
            "no-focus",
            "auto-close",
            "position",
            "full-height",
            "self",
            "maximized",
            "offset",
            "anchor",
            "no-refocus",
            "target",
            "persistent",
        ]
        self.allowed_events = [
            "input",
            "show",
            "before_show",
            "hide",
            "before_hide",
            "escape_key",
        ]
        self.set_keyword_events(**kwargs)


@parse_dict
class QDialog(_QInputBase):
    slots = []
    html_tag = "q-dialog"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.type = "boolean"
        self.value = bool(self.value)
        self.prop_list = [
            "persistent",
            "no-esc-dsimiss",
            "no-backdrop-dismiss",
            "no-route-dismiss",
            "auto-close",
            "transition-show",
            "transition-hide",
            "no-refocus",
            "no-focus",
            "seamless",
            "maximized",
            "full-width",
            "full-height",
            "position",
            "value",
            "content-class",
            "content-style",
            "square",
        ]
        self.allowed_events = [
            "input",
            "show",
            "before_show",
            "hide",
            "before_hide",
            "escape_key",
        ]
        self.set_keyword_events(**kwargs)


@parse_dict
class QTooltip(_QInputBase):
    slots = ["default_slot"]
    html_tag = "q-tooltip"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.disable_events = True  # For tooltips, events are disabled by default otherwise input event occurs every time tooltip shows.
        self.type = "boolean"
        self.value = bool(self.value)
        self.prop_list = [
            "transition-show",
            "transition-hide",
            "target",
            "delay",
            "max-height",
            "max-width",
            "value",
            "anchor",
            "self",
            "offset",
            "content-class",
            "content-style",
            "hide-delay",
        ]
        self.allowed_events = [
            "input",
            "show",
            "before_show",
            "hide",
            "before_hide",
            "before",
        ]
        self.set_keyword_events(**kwargs)


@parse_dict
class QStepper(_QInputBase):
    # Not working
    html_tag = "q-stepper"

    slots = ["default_slot", "navigation_slot", "message_slot"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "keep-alive",
            "animated",
            "infinite",
            "swipeable",
            "transition-prev",
            "transition-next",
            "vertical",
            "header-nav",
            "value",
            "dark",
            "flat",
            "bordered",
            "alternative-labels",
            "contracted",
            "inactive-icon",
            "inactive-color",
            "done-icon",
            "done-color",
            "active-icon",
            "active-color",
            "error-icon",
            "error-color",
        ]
        self.allowed_events = ["input", "before-transition", "transition"]
        self.set_keyword_events(**kwargs)


@parse_dict
class QStep(QDiv):
    slots = ["default_slot"]
    html_tag = "q-step"

    def __init__(self, **kwargs):
        self.done = False
        super().__init__(**kwargs)
        self.prop_list = [
            "header-nav",
            "name",
            "disable",
            "done",
            "error",
            "color",
            "icon",
            "title",
            "caption",
            "prefix",
            "done-icon",
            "done-color",
            "active-icon",
            "active-color",
            "error-icon",
            "error-color",
        ]


@parse_dict
class QStepperNavigation(QDiv):
    slots = ["default_slot"]
    html_tag = "q-stepper-navigation"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = []


@parse_dict
class QSlideTransition(QDiv):
    slots = ["default_slot"]
    html_tag = "q-slide-transition"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["appear", "duration"]


@parse_dict
class QTimeline(QDiv):
    slots = ["default_slot"]
    html_tag = "q-timeline"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["side", "layout", "color", "dark"]


@parse_dict
class QTimelineEntry(QDiv):
    slots = ["default_slot", "title_slot", "subtitle_slot"]
    html_tag = "q-timeline-entry"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "color",
            "heading",
            "tag",
            "icon",
            "avatar",
            "title",
            "subtitle",
            "body",
            "side",
        ]


@parse_dict
class QEditor(QInput):
    slots = []
    html_tag = "q-editor"

    icon_set = Dict(
        {
            "align": "format_align_left",
            "bold": "format_bold",
            "center": "format_align_center",
            "code": "code",
            "font": "font_download",
            "fontSize": "format_size",
            "formatting": "text_format",
            "header": "format_size",
            "hr": "remove",
            "hyperlink": "link",
            "indent": "format_indent_increase",
            "italic": "format_italic",
            "justify": "format_align_justify",
            "left": "format_align_left",
            "orderedList": "format_list_numbered",
            "outdent": "format_indent_decrease",
            "print": "print",
            "quote": "format_quote",
            "redo": "redo",
            "removeFormat": "format_clear",
            "right": "format_align_right",
            "size": "format_size",
            "strikethrough": "strikethrough_s",
            "subscript": "vertical_align_bottom",
            "superscript": "vertical_align_top",
            "toggleFullscreen": "fullscreen",
            "underline": "format_underlined",
            "undo": "undo",
            "unorderedList": "format_list_bulleted",
        }
    )

    lang = Dict(
        {
            "url": "URL",
            "bold": "Bold",
            "italic": "Italic",
            "strikethrough": "Strikethrough",
            "underline": "Underline",
            "unorderedList": "Unordered List",
            "orderedList": "Ordered List",
            "subscript": "Subscript",
            "superscript": "Superscript",
            "hyperlink": "Hyperlink",
            "toggleFullscreen": "Toggle Fullscreen",
            "quote": "Quote",
            "left": "Left align",
            "center": "Center align",
            "right": "Right align",
            "justify": "Justify align",
            "print": "Print",
            "outdent": "Decrease indentation",
            "indent": "Increase indentation",
            "removeFormat": "Remove formatting",
            "formatting": "Formatting",
            "fontSize": "Font Size",
            "align": "Align",
            "hr": "Insert Horizontal Rule",
            "undo": "Undo",
            "redo": "Redo",
            "header1": "Header 1",
            "header2": "Header 2",
            "header3": "Header 3",
            "header4": "Header 4",
            "header5": "Header 5",
            "header6": "Header 6",
            "paragraph": "Paragraph",
            "code": "Code",
            "size1": "Very small",
            "size2": "A bit small",
            "size3": "Normal",
            "size4": "Medium-large",
            "size5": "Big",
            "size6": "Very big",
            "size7": "Maximum",
            "defaultFont": "Default Font",
        }
    )

    simple_options = [
        ["left", "center", "right", "justify"],
        ["bold", "italic", "strike", "underline", "subscript", "superscript"],
        ["hr", "link"],
        ["undo", "redo"],
        ["print", "fullscreen"],
    ]

    fonts = {
        "arial": "Arial",
        "arial_black": "Arial Black",
        "comic_sans": "Comic Sans MS",
        "courier_new": "Courier New",
        "impact": "Impact",
        "lucida_grande": "Lucida Grande",
        "times_new_roman": "Times New Roman",
        "verdana": "Verdana",
    }

    kitchen_sink = [
        [
            {
                "label": lang.align,
                "icon": icon_set.align,
                "fixedLabel": True,
                "list": "only-icons",
                "options": ["left", "center", "right", "justify"],
            },
            {
                "label": lang.align,
                "icon": icon_set.align,
                "fixedLabel": True,
                "options": ["left", "center", "right", "justify"],
            },
        ],
        ["bold", "italic", "strike", "underline", "subscript", "superscript"],
        ["token", "hr", "link", "custom_btn"],
        ["print", "fullscreen"],
        [
            {
                "label": lang.formatting,
                "icon": icon_set.formatting,
                "list": "no-icons",
                "options": ["p", "h1", "h2", "h3", "h4", "h5", "h6", "code"],
            },
            {
                "label": lang.fontSize,
                "icon": icon_set.fontSize,
                "fixedLabel": True,
                "fixedIcon": True,
                "list": "no-icons",
                "options": [
                    "size-1",
                    "size-2",
                    "size-3",
                    "size-4",
                    "size-5",
                    "size-6",
                    "size-7",
                ],
            },
            {
                "label": lang.defaultFont,
                "icon": icon_set.font,
                "fixedIcon": True,
                "list": "no-icons",
                "options": [
                    "default_font",
                    "arial",
                    "arial_black",
                    "comic_sans",
                    "courier_new",
                    "impact",
                    "lucida_grande",
                    "times_new_roman",
                    "verdana",
                ],
            },
            "removeFormat",
        ],
        ["quote", "unordered", "ordered"],
        ["undo", "redo"],
    ]

    def __init__(self, **kwargs):
        self.kitchen_sink = False
        self.toolbar = QEditor.simple_options
        super().__init__(**kwargs)
        self.prop_list = [
            "fullscreen",
            "value",
            "readonly",
            "disable",
            "square",
            "flat",
            "dense",
            "min-height",
            "max-height",
            "height",
            "toolbar-outline",
            "toolbar-push",
            "toolbar-rounded",
            "content-style",
            "content-class",
            "definitions",
            "fonts",
            "toolbar",
            "toolbar-color",
            "toolbar-text-color",
            "toolbar-toggle-color",
            "toolbar-bg",
            "toolbar-outline",
            "toolbar-push",
            "toolbar-rounded",
        ]
        self.allowed_events = ["input"]
        # self.set_keyword_events(**kwargs)

    def convert_object_to_dict(self):
        self.debounce = 0  # Component has its own debounce mechanism
        if self.kitchen_sink:
            self.toolbar = QEditor.kitchen_sink
            self.fonts = QEditor.fonts
        d = super().convert_object_to_dict()
        return d


@parse_dict
class QExpansionItem(_QInputBase):
    slots = ["default_slot", "header_slot"]
    html_tag = "q-expansion-item"

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.type = "boolean"
        self.value = bool(self.value)
        self.prop_list = [
            "to",
            "exact",
            "append",
            "replace",
            "active-class",
            "exact-active-class",
            "duration",
            # 'default-opened',  Do not use, just set value to True: self.value = True
            "expand-icon-toggle",
            "group",
            "popup",
            "icon",
            "expand-icon",
            "label",
            "label-lines",
            "caption",
            "caption-lines",
            "header-inset-level",
            "content-inset-level",
            "expand-separator",
            "switch-toggle-side",
            "value",
            "disable",
            "expand-icon-class",
            "dark",
            "dense",
            "dense-toggle",
            "header-style",
            "header-class",
        ]
        self.allowed_events = [
            "input",
            "show",
            "before_show",
            "hide",
            "before_hide",
            "escape_key",
        ]
        self.set_keyword_events(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        try:
            if d["attrs"]["default-opened"]:
                d["value"] = True
                d["attrs"]["value"] = True
        except:
            pass
        return d


@parse_dict
class QImg(QDiv):
    slots = ["defualt_slot", "loading_slot", "error_slot"]
    html_tag = "q-img"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "transition",
            "alt",
            "basic",
            "contain",
            "position",
            "ratio",
            "src",
            "srcset",
            "sizes",
            "placeholder-src",
            "spinner-color",
            "spinner-size",
        ]
        # Events: load, error


@parse_dict
class QInnerLoading(QDiv):
    slots = ["default_slot"]
    html_tag = "q-inner-loading"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "transition-show",
            "transition-hide",
            "showing",
            "color",
            "size",
            "dark",
        ]


@parse_dict
class QParallax(QDiv):
    slots = ["default_slot", "media_slot", "content_slot"]
    html_tag = "q-parallax"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["speed", "src", "height"]


@parse_dict
class Transition(QDiv):
    html_tag = "transition"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "appear",
            "appear-class",
            "appear-to-class",
            "appear-active-class",
            "enter-class",
            "enter-active-class",
            "enter-to-class",
            "leave-class",
            "leave-active-class",
            "leave-to-class",
            "duration",
            "mode",
        ]


@parse_dict
class TransitionGroup(QDiv):
    html_tag = "transition-group"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "appear",
            "appear-class",
            "appear-to-class",
            "appear-active-class",
            "enter-class",
            "enter-active-class",
            "enter-to-class",
            "leave-class",
            "leave-active-class",
            "leave-to-class",
            "duration",
            "mode",
            "tag",
            "key",
        ]


@parse_dict
class QIcon(QDiv):
    slots = ["default_slot"]
    html_tag = "q-icon"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["name", "color", "size", "left", "right", "tag"]


@parse_dict
class QSpinner(QDiv):
    # TODO: Take care of all spinner types in parser

    spinner_types = [
        "audio",
        "ball",
        "bars",
        "comment",
        "cube",
        "dots",
        "facebook",
        "gears",
        "grid",
        "hearts",
        "hourglass",
        "infinity",
        "ios",
        "oval",
        "pie",
        "puff",
        "radio",
        "rings",
        "tail",
    ]
    html_tag = "q-spinner"

    def __init__(self, **kwargs):

        self.size = "1em"
        self.color = "primary"
        self.spinner_type = ""
        super().__init__(**kwargs)
        self.prop_list = ["size", "color", "thickness"]

    def convert_object_to_dict(self):
        if self.spinner_type in QSpinner.spinner_types:
            self.html_tag = "q-spinner-" + self.spinner_type
        else:
            self.html_tag = "q-spinner"
        d = super().convert_object_to_dict()
        return d


@parse_dict
class QList(QDiv):
    slots = ["default_slot"]
    html_tag = "q-list"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["separator", "padding", "bordered", "dense", "dark"]


@parse_dict
class QItem(QDiv):
    slots = ["default_slot"]
    html_tag = "q-item"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "to",
            "exact",
            "append",
            "replace",
            "active-class",
            "exact-active-class",
            "inset-level",
            "tag",
            "tabindex",
            "disable",
            "active",
            "clickable",
            "manual-focus",
            "focused",
            "dark",
            "dense",
        ]


@parse_dict
class QItemSection(QDiv):
    slots = ["default_slot"]
    html_tag = "q-item-section"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["avatar", "thumbnail", "side", "top", "no-wrap"]


@parse_dict
class QItemLabel(QDiv):
    slots = ["default_slot"]
    html_tag = "q-item-label"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["lines", "overline", "caption", "header"]


@parse_dict
class QSlideItem(QDiv):
    slots = ["default_slot", "left_slot", "right_slot"]
    html_tag = "q-slide-item"

    def __init__(self, **kwargs):
        self.reset = False
        super().__init__(**kwargs)
        self.prop_list = ["left-color", "right-color"]
        self.allowed_events = ["left", "right", "action", "bottom", "top"]
        self.set_keyword_events(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["reset"] = self.reset
        return d


@parse_dict
class QInfiniteScroll(QDiv):
    # https://quasar.dev/vue-components/infinite-scroll

    # To work MUST add 'scroll' to classes of container that holds this component. Otherwise endless load events generated
    # Also styling must be through style not color or size props because they disappear after first event

    slots = ["default_slot", "loading_slot"]
    html_tag = "q-infinite-scroll"

    def __init__(self, **kwargs):
        self.done = False
        self.initial = True
        super().__init__(**kwargs)
        self.prop_list = ["offset", "scrollTarget", "reverse", "disable"]
        self.allowed_events = ["load"]
        self.set_keyword_events(**kwargs)

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["done"] = self.done
        return d


@parse_dict
class QScrollArea(QDiv):
    slots = ["default_slot"]
    html_tag = "q-scroll-area"

    def __init__(self, **kwargs):
        self.offset = None
        self.duration = None
        super().__init__(**kwargs)
        # Offset and duration are added props to activate the setScrollPosition function
        self.prop_list = [
            "delay",
            "horizontal",
            "thumb-style",
            "content-style",
            "content-active-style",
            "offset",
            "duration",
        ]

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["offset"] = self.offset
        d["duration"] = self.duration
        return d


@parse_dict
class QScrollObserver(QDiv):
    html_tag = "q-scroll-observer"

    def __init__(self, **kwargs):
        self.debounce = 300
        super().__init__(**kwargs)
        self.prop_list = ["debounce", "horizontal"]
        self.allowed_events = ["scroll"]
        self.set_keyword_events(**kwargs)


@parse_dict
class QMarkupTable(QDiv):
    slots = ["default_slot"]
    html_tag = "q-markup-table"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "separator",
            "wrap-cells",
            "dense",
            "dark",
            "flat",
            "bordered",
            "square",
        ]


@parse_dict
class QSeparator(QDiv):
    html_tag = "q-separator"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["spaced", "inset", "vertical", "dark", "color"]


@parse_dict
class QSpace(QDiv):
    html_tag = "q-space"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = []


@parse_dict
class QVideo(QDiv):
    html_tag = "q-video"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["src", "ratio"]


@parse_dict
class QTree(QDiv):
    slots = []
    html_tag = "q-tree"

    def __init__(self, **kwargs):

        self.ticked = []
        self.expanded = []
        self.selected = []
        self.tick_strategy = "none"  # none | strict | leaf | leaf-filtered
        self.default_expand_all = False
        super().__init__(**kwargs)
        self.attributes = [
            "tick-strategy",
            "default-expand-all",
            "accordion",
            "nodes",
            "node-key",
            "label-key",
            "icon",
            "no-nodes-label",
            "no-results-label",
            "filter",
            "filter-method",
            "ticked",
            "expanded",
            "selected",
            "color",
            "control-color",
            "text-color",
            "selected-color",
            "dark",
            "duration",
            "no-connectors",
        ]
        self.allowed_events = [
            "before",
            "after",
            "update:expanded",
            "lazy-load",
            "update:ticked",
            "update:selected",
        ]
        self.set_keyword_events(**kwargs)
        self.events = ["update:expanded", "update:ticked", "update:selected"]

        def default_input(self, msg):
            return self.before_event_handler(msg)

        self.on("before", default_input)

    def before_event_handler(self, msg):
        if msg.event_type == "update:expanded":
            self.expanded = msg.value
        elif msg.event_type == "update:selected":
            self.selected = msg.value
        elif msg.event_type == "update:ticked":
            self.ticked = msg.value

    def model_update(self):
        pass

    def __setattr__(self, key, value):
        if key == "nodes":
            if isinstance(value, str):
                self.load_json(value)
            else:
                self.__dict__[key] = value
        elif key in self.slots:
            self.add_scoped_slot(key[: key.index("_")], value)
        else:
            self.__dict__[key] = value

    def load_json(self, options_string):
        self.nodes = hjson.loads(options_string.encode("ascii", "ignore"))
        return self.nodes

    def load_json_from_file(self, file_name):
        with open(file_name, "r") as f:
            self.nodes = hjson.loads(f.read().encode("ascii", "ignore"))
        return self.nodes

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["default_expand_all"] = self.default_expand_all
        return d


@parse_dict
class QNotify(QDiv):
    # TODO: Add handler for action that sends message back to server. Currently does not support actions

    html_tag = "q-notify"

    # This is a special component that does not wrap a quasar component but activates a quasar utility
    # https://quasar.dev/quasar-plugins/notify
    # Add 'after' event to button that activates that sets 'notify' prop back to false. Otherwise, will appear every update
    # Or, await page update in event and then set notify to false

    def __init__(self, **kwargs):
        self.notify = False
        # Possible position: top-left | top-right | bottom-left | bottom-right | top | bottom | left | right | center
        self.position = "bottom-right"
        super().__init__(**kwargs)
        self.prop_list = [
            "notify",
            "color",
            "textColor",
            "message",
            "html",
            "icon",
            "avatar",
            "position",
            "closeBtn",
            "timeout",
            "actions",
            "multiLine",
            "caption",
            "reply",
        ]

    def convert_object_to_dict(self):
        d = super().convert_object_to_dict()
        d["notify"] = self.notify
        return d


@parse_dict
class QTable(QDiv):
    # Has also specific slots for column names body-cell-{name}
    slots = [
        "item_slot",
        "body_slot",
        "body-cell_slot",
        "top-row_slot",
        "bottom-row_slot",
        "bottom_slot",
        "pagination_slot",
        "header_slot",
        "header-cell_slot",
        "top-left_slot",
        "top-right_slot",
        "top-selection_slot",
        "top_slot",
        "bottom_slot",
        "loading_slot",
    ]
    html_tag = "q-table"

    def __init__(self, **kwargs):

        self.pagination = False
        self.selected = []
        super().__init__(**kwargs)
        self.attributes = [
            "fullscreen",
            "loading",
            "columns",
            "visible-columns",
            "title",
            "hide-header",
            "hide-bottom",
            "separator",
            "wrap-cells",
            "no-data-label",
            "no-results-label",
            "loading-label",
            "filter",
            "filter-method",
            "data",
            "row-key",
            "rows-per-page-label",
            "pagination-label",
            "pagination",
            "rows-per-page-options",
            "selected-rows-label",
            "selection",
            "selected",
            "binary-state-sort",
            "sort-method",
            "color",
            "grid",
            "dense",
            "dark",
            "flat",
            "bordered",
            "square",
            "table-style",
            "table-class",
            "table-header-style",
            "table-header-class",
            "card-style",
            "card-class",
            "icon-first-page",
            "con-prev-page",
            "icon-next-page",
            "icon-last-page",
        ]
        # Need to load: columns, data
        # separator one of  horizontal | vertical | cell | none
        self.allowed_events = [
            "before",
            "after",
            "request",
            "selection",
            "update:pagination",
            "update:selected",
        ]
        self.set_keyword_events(**kwargs)
        self.events = ["update:pagination", "update:selected"]

        def default_input(self, msg):
            return self.before_event_handler(msg)

        self.on("before", default_input)

    def before_event_handler(self, msg):
        if msg.event_type == "update:selected":
            self.selected = msg.value
        elif msg.event_type == "update:pagination":
            self.pagination = msg.value

    def model_update(self):
        pass

    def __setattr__(self, key, value):
        if key in ["data", "columns"]:
            if isinstance(value, str):
                self.__dict__[key] = self.load_json(value)
            else:
                self.__dict__[key] = value
        elif key in self.slots:
            self.add_scoped_slot(key[: key.index("_")], value)
        else:
            self.__dict__[key] = value

    def load_json(self, options_string):
        self.nodes = hjson.loads(options_string.encode("ascii", "ignore"))
        return self.nodes

    def load_json_from_file(self, file_name):
        with open(file_name, "r") as f:
            self.nodes = hjson.loads(f.read().encode("ascii", "ignore"))
        return self.nodes

    def load_pandas_frame(self, df):
        self.columns = [
            Dict(
                {
                    "name": col,
                    "align": "center",
                    "label": col,
                    "field": col,
                    "sortable": True,
                }
            )
            for col in df.columns
        ]
        self.data = df.to_dict("records")


@parse_dict
class QTh(QDiv):
    slots = ["default_slot"]
    html_tag = "q-th"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # position one of top-right | top-left | bottom-right | bottom-left | top | right | bottom | left
        self.prop_list = ["props", "auto-width"]
        self.allowed_events = []
        self.set_keyword_events(**kwargs)


@parse_dict
class QTr(QDiv):
    slots = ["default_slot"]
    html_tag = "q-tr"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # position one of top-right | top-left | bottom-right | bottom-left | top | right | bottom | left
        self.prop_list = ["props", "no-hover"]
        self.allowed_events = []
        self.set_keyword_events(**kwargs)


@parse_dict
class QTd(QDiv):
    slots = ["default_slot"]
    html_tag = "q-td"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # position one of top-right | top-left | bottom-right | bottom-left | top | right | bottom | left
        self.prop_list = ["auto-width", "props", "no-hover"]
        self.allowed_events = []
        self.set_keyword_events(**kwargs)


# Layout components  https://quasar.dev/layout/layout


@parse_dict
class QLayout(QDiv):
    # To understand the 'view' prop read https://quasar.dev/layout-builder

    slots = ["default_slot"]
    html_tag = "q-layout"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # position one of top-right | top-left | bottom-right | bottom-left | top | right | bottom | left
        self.prop_list = ["view", "container"]
        self.allowed_events = ["resize", "scroll", "scroll-height"]
        self.set_keyword_events(**kwargs)


@parse_dict
class QHeader(QDiv):
    slots = ["default_slot"]
    html_tag = "q-header"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # position one of top-right | top-left | bottom-right | bottom-left | top | right | bottom | left
        self.prop_list = ["reveal", "reveal-offset", "value", "bordered", "elevated"]
        self.allowed_events = ["reveal"]
        self.set_keyword_events(**kwargs)


@parse_dict
class QFooter(QDiv):
    slots = ["default_slot"]
    html_tag = "q-footer"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["reveal", "reveal-offset", "value", "bordered", "elevated"]
        self.allowed_events = ["reveal"]
        self.set_keyword_events(**kwargs)


@parse_dict
class QDrawer(QDiv):
    slots = ["default_slot", "mini_slot"]
    html_tag = "q-drawer"

    def __init__(self, **kwargs):
        self.value = True  # Default is to show drawer
        self.mini = False
        super().__init__(**kwargs)
        self.prop_list = [
            "side",
            "overlay",
            "mini",
            "mini-to-overlay",
            "breakpoint",
            "behavior",
            "persistent",
            "show-if-above",
            "no-swipe-open",
            "no-swipe-close",
            "value",
            "width",
            "mini-width",
            "bordered",
            "elevated",
            "content-class",
            "content-style",
        ]
        self.allowed_events = [
            "before",
            "input",
            "show",
            "before-show",
            "hide",
            "before-hide",
            "on-layout",
            "click",
            "mouseover",
            "mouseout",
        ]
        self.set_keyword_events(**kwargs)
        self.on("input", self.default_input)
        self.on("before", self.default_before)

    def default_before(self, msg):
        if hasattr(self, "model"):
            self.set_model(msg.value)
        self.value = msg.value

    def default_input(self, msg):
        pass

    async def toggle(self, msg):
        await self.run_method("toggle()", msg.websocket)

    def model_update(self):
        update_value = self.model[0].data[self.model[1]]
        self.value = update_value


@parse_dict
class QPageContainer(QDiv):
    slots = ["default_slot"]
    html_tag = "q-page-container"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = []


@parse_dict
class QPage(QDiv):
    slots = ["default_slot"]
    html_tag = "q-page"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = ["padding", "style-fn"]


@parse_dict
class QPageSticky(QDiv):
    slots = ["default_slot"]
    html_tag = "q-page-sticky"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # position one of top-right | top-left | bottom-right | bottom-left | top | right | bottom | left
        self.prop_list = ["expand", "position", "offset"]


@parse_dict
class QPageScroller(QDiv):
    slots = ["default_slot"]
    html_tag = "q-page-scroller"

    def __init__(self, **kwargs):
        self.duration = 300
        super().__init__(**kwargs)
        # position one of top-right | top-left | bottom-right | bottom-left | top | right | bottom | left
        self.prop_list = ["expand", "position", "offset", "scroll-offset", "duration"]
        # Required in order not have the vue update during the scroll
        self.on("click", self.default_click)

    @staticmethod
    async def default_click(self, msg):
        await asyncio.sleep(1 + self.duration / 1000)


@parse_dict
class QFab(QDiv):
    slots = ["default_slot", "tooltip_slot"]
    html_tag = "q-fab"

    def __init__(self, **kwargs):
        self.value = False  # Default is to show drawer
        super().__init__(**kwargs)
        # direction on of  up | right | down | left   , type one of  a | submit | button | reset
        self.prop_list = [
            "direction",
            "persistent",
            "icon",
            "active-icon",
            "type",
            "value",
            "disable",
            "outline",
            "push",
            "flat",
            "color",
            "text-color",
            "glossy",
            "label",
            "external-label",
            "label-position",
        ]
        self.allowed_events = [
            "input",
            "show",
            "before-show",
            "hide",
            "before-hide",
            "click",
            "mouseover",
            "mouseout",
        ]
        self.set_keyword_events(**kwargs)

    def model_update(self):
        update_value = self.model[0].data[self.model[1]]
        self.value = update_value


@parse_dict
class QFabAction(QDiv):
    slots = ["default_slot"]
    html_tag = "q-fab-action"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # position one of top-right | top-left | bottom-right | bottom-left | top | right | bottom | left
        self.prop_list = [
            "icon",
            "type",
            "to",
            "replace",
            "disable",
            "outline",
            "push",
            "flat",
            "color",
            "text-color",
            "glossy",
            "label",
            "external-label",
            "label-position",
        ]


@parse_dict
class QSkeleton(QDiv):
    slots = ["default_slot"]
    html_tag = "q-skeleton"

    def __init__(self, **kwargs):
        self.type = "rect"
        self.tag = "div"
        super().__init__(**kwargs)
        if not self.animation:
            self.animation = "wave"
        self.prop_list = [
            "tag",
            "type",
            "dark",
            "animation",
            "square",
            "bordered",
            "size",
            "width",
            "height",
        ]


@parse_dict
class QPopupEdit(_QInputBase):
    # Work in progress
    # 'cancel' and 'save' events not working
    html_tag = "q-popup-edit"
    slots = ["default_slot", "title_slot"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.prop_list = [
            "touch-position",
            "persistent",
            "separate-close-popup",
            "title",
            "buttons",
            "label-set",
            "label-cancel",
            "value",
            "validate",
            "fit",
            "cover",
            "anchor",
            "self",
            "disable",
            "content-class",
            "content-style",
            "color",
            "offset",
            "square",
            "max-height",
            "max-width",
        ]

        self.allowed_events = ["input", "cancel", "save"]
        self.set_keyword_events(**kwargs)

        def default_event_handler(self, msg):
            print(msg)

        self.on("save", default_event_handler)
        self.on("cancel", default_event_handler)


class ToggleDarkModeBtn(QBtn):
    def __init__(self, **kwargs):
        self.label = "Toggle Dark Mode"
        super().__init__(**kwargs)
        self.on("click", self.toggle_dark_mode)

    @staticmethod
    async def toggle_dark_mode(self, msg):
        msg.page.dark = not msg.page.dark
        await msg.page.set_dark_mode(msg.page.dark)


class QInputDateTime(QInput):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.mask = "####-##-## ##:##"
        self.fill_mask = True
        self.proxy = {}

        date_slot = QIcon(name="event", classes="cursor-pointer")
        c2 = QPopupProxy(transition_show="scale", transition_hide="scale", a=date_slot)
        self.date = QDate(mask="YYYY-MM-DD HH:mm", name="date", a=c2)
        self.proxy["QDate"] = c2

        time_slot = QIcon(name="access_time", classes="cursor-pointer")
        c2 = QPopupProxy(transition_show="scale", transition_hide="scale", a=time_slot)
        self.time = QTime(mask="YYYY-MM-DD HH:mm", format24h=True, name="time", a=c2)
        self.proxy["QTime"] = c2

        self.date.parent = self
        self.time.parent = self
        self.date.value = self.value
        self.time.value = self.value
        self.date.model = self.model
        self.time.model = self.model
        self.prepend_slot = date_slot
        self.append_slot = time_slot
        self.date.on("input", self.date_time_change)
        self.time.on("input", self.date_time_change)
        self.on("input", self.input_change)

    @staticmethod
    async def date_time_change(self, msg):
        self.parent.value = self.value
        self.parent.date.value = self.value
        self.parent.time.value = self.value
        await self.parent.proxy[msg.class_name].run_method("hide()", msg.websocket)

    @staticmethod
    def input_change(self, msg):
        self.date.value = self.value
        self.time.value = self.value


class QInputDate(QInput):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        date_slot = QIcon(name="event", classes="cursor-pointer")
        c2 = QPopupProxy(transition_show="scale", transition_hide="scale", a=date_slot)
        self.date = QDate(mask="YYYY-MM-DD", name="date", a=c2)

        self.date.parent = self
        self.date.value = self.value
        self.append_slot = date_slot
        self.date.on("input", self.date_time_change)
        self.on("input", self.input_change)
        self.proxy = c2

    @staticmethod
    async def date_time_change(self, msg):
        self.parent.value = self.value
        self.parent.date.value = self.value
        await self.parent.proxy.run_method("hide()", msg.websocket)

    @staticmethod
    def input_change(self, msg):
        self.date.value = self.value


@parse_dict
class QResizeObserver(QDiv):
    html_tag = "q-resize-observer"

    def __init__(self, **kwargs):
        self.debounce = 300
        super().__init__(**kwargs)
        self.prop_list = ["debounce"]
        self.allowed_events = ["resize"]
        self.set_keyword_events(**kwargs)



# Justpy Tutorial demo hello_test1 from docs/tutorial/custom_components.md
import justpy as jp

def hello_test1():
    wp = jp.WebPage()
    h = jp.Hello()
    for i in range(5):
        wp.add(h)
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("hello_test1",hello_test1)



'''
Created on 2022-09-13

@author: wf
'''
import justpy as jp
from tests.basetest import Basetest
from starlette.testclient import TestClient

class BaseClienttest(Basetest):
    """
    Basetests using the starlette TestClient
    """
    
    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.app = jp.app
    
    def checkResponse(self, path, expected_code=200, debug: bool = None):
        """
        check the response for the given path

        Args:
            path(str): the path to check
            expected_code(int): the HTTP status code to expect
            debug(bool): if True show debugging info
        """
        if debug is None:
            debug = self.debug
        if debug:
            print(f"checking response for {path}")
        with TestClient(self.app) as client:
            response = client.get(path)
            self.assertEqual(expected_code, response.status_code)
        if debug:
            print(response.text)
        return response



# Justpy Tutorial demo clock_test from docs/tutorial/pushing_data.md
import justpy as jp
import time
import asyncio

wp = jp.WebPage(delete_flag=False)
clock_div = jp.Span(text='Loading...', classes='text-5xl m-1 p-1 bg-gray-300 font-mono', a=wp)

async def clock_counter():
    while True:
        clock_div.text = time.strftime("%a, %d %b %Y, %H:%M:%S", time.localtime())
        jp.run_task(wp.update())
        await asyncio.sleep(1)

async def clock_init():
    jp.run_task(clock_counter())

async def clock_test():
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("clock_test",clock_test, startup=clock_init)



# Justpy Tutorial demo demo_function from docs/tutorial/request_object.md
import justpy as jp

def demo_function(request):
    wp = jp.WebPage()
    if len(request.query_params) > 0:
        for key, value in request.query_params.items():
            jp.P(text=f'{key}: {value}', a=wp, classes='text-xl m-2 p-1')
    else:
        jp.P(text='No URL paramaters present', a=wp, classes='text-xl m-2 p-1')
    return wp

# initialize the demo
from  examples.basedemo import Demo
Demo ("demo_function",demo_function)
